from django.urls import path

from . import views

app_name = 'monitoring_general'

urlpatterns = [
    path(
        'api/v1/monitoring/dashboard/',
        views.dashboard_timeseries,
        name='api_dashboard_timeseries',
    ),
]



# Manually created

import logging

from django.db import migrations

from .influxdb.influxdb_alter_structure_0006 import (
    update_metric_timeseries_structure_forward_migration,
    update_metric_timeseries_structure_reverse_migration,
)

CHUNK_SIZE = 1000

logger = logging.getLogger(__name__)


class Migration(migrations.Migration):
    dependencies = [('monitoring', '0004_metric_main_and_extra_tags')]

    operations = [
        migrations.RunPython(
            update_metric_timeseries_structure_forward_migration,
            reverse_code=update_metric_timeseries_structure_reverse_migration,
        )
    ]



import uuid
from unittest.mock import patch

from django.test import TestCase
from django.urls import reverse
from swapper import load_model

from openwisp_controller.config.tests.utils import CreateConfigTemplateMixin
from openwisp_controller.geo.tests.utils import TestGeoMixin

from ..configuration import DEFAULT_DASHBOARD_TRAFFIC_CHART
from . import TestMonitoringMixin

Organization = load_model('openwisp_users', 'Organization')
Metric = load_model('monitoring', 'Metric')
OrganizationUser = load_model('openwisp_users', 'OrganizationUser')
Group = load_model('openwisp_users', 'Group')
DeviceLocation = load_model('geo', 'DeviceLocation')
FloorPlan = load_model('geo', 'FloorPlan')
Location = load_model('geo', 'Location')
Device = load_model('config', 'Device')


@patch.dict(DEFAULT_DASHBOARD_TRAFFIC_CHART, {'__all__': ['wan']})
class TestDashboardTimeseriesView(
    CreateConfigTemplateMixin, TestMonitoringMixin, TestGeoMixin, TestCase
):
    location_model = Location
    floorplan_model = FloorPlan
    org2_id = str(uuid.uuid4())
    org3_id = str(uuid.uuid4())

    def _test_response_data(self, response):
        self.assertIn('x', response.data)
        charts = response.data['charts']
        for chart in charts:
            self.assertIn('traces', chart)
            self.assertIn('title', chart)
            self.assertIn('description', chart)
            self.assertIn('type', chart)
            self.assertIn('summary', chart)
            self.assertIsInstance(chart['summary'], dict)
            self.assertIn('summary_labels', chart)
            self.assertIsInstance(chart['summary_labels'], list)
            self.assertIn('unit', chart)
            self.assertIn('colors', chart)
            self.assertIn('colorscale', chart)

    def _create_org_wifi_client_metric(self, org, **kwargs):
        extra_tags = {'organization_id': str(org.id)}
        extra_tags.update(kwargs)
        return self._create_general_metric(
            name='wifi_clients',
            configuration='general_clients',
            field_name='clients',
            main_tags={'ifname': 'wlan0'},
            extra_tags=extra_tags,
        )

    def _create_org_traffic_metric(self, org, interface_name):
        return self._create_general_metric(
            name='traffic',
            configuration='general_traffic',
            field_name='rx_bytes',
            main_tags={'ifname': interface_name},
            extra_tags={'organization_id': str(org.id)},
        )

    def _create_test_users(self, org):
        admin = self._create_admin()
        org_admin = self._create_user()
        OrganizationUser.objects.create(user=org_admin, organization=org, is_admin=True)
        groups = Group.objects.filter(name='Administrator')
        org_admin.groups.set(groups)
        operator = self._create_operator()
        return admin, org_admin, operator

    def test_wifi_client_chart(self):
        def _test_chart_properties(chart):
            self.assertEqual(chart['title'], 'General WiFi Clients')
            self.assertEqual(chart['type'], 'bar')
            self.assertEqual(chart['unit'], '')
            self.assertEqual(chart['summary_labels'], ['Total Unique WiFi clients'])
            self.assertEqual(chart['colors'], ['#1f77b4'])
            self.assertEqual(chart['colorscale'], None)
            self.assertEqual(
                chart['description'],
                'Unique WiFi clients count of the entire network.',
            )

        path = reverse('monitoring_general:api_dashboard_timeseries')
        org1 = self._create_org(name='org1', slug='org1')
        org2 = self._create_org(name='org2', slug='org2')
        org3 = self._create_org(name='org3', slug='org3')
        org1_metric = self._create_org_wifi_client_metric(org1)
        org2_metric = self._create_org_wifi_client_metric(org2)
        org3_metric = self._create_org_wifi_client_metric(org3)

        org1_metric.write('00:23:4a:00:00:00')
        org2_metric.write('00:23:4b:00:00:00')
        org3_metric.write('00:14:5c:00:00:00')

        admin, org2_administrator, operator = self._create_test_users(org2)
        OrganizationUser.objects.create(
            user=org2_administrator, organization=org3, is_admin=True
        )

        self.client.force_login(admin)
        with self.subTest('Test superuser retrieves metric for all organizations'):
            response = self.client.get(path)
            self.assertEqual(response.status_code, 200)
            self._test_response_data(response)
            chart = response.data['charts'][0]
            self.assertEqual(chart['traces'][0][1][-1], 3)
            self.assertEqual(chart['summary']['wifi_clients'], 3)
            _test_chart_properties(chart)

        with self.subTest('Test superuser retrieves metric for one organization'):
            response = self.client.get(path, {'organization_slug': org2.slug})
            self.assertEqual(response.status_code, 200)
            self._test_response_data(response)
            chart = response.data['charts'][0]
            self.assertEqual(chart['traces'][0][1][-1], 1)
            self.assertEqual(chart['summary']['wifi_clients'], 1)
            _test_chart_properties(chart)

        with self.subTest('Test superuser retrieves metric for multiple organization'):
            response = self.client.get(
                path, {'organization_slug': f'{org1.slug},{org2.slug}'}
            )
            self.assertEqual(response.status_code, 200)
            self._test_response_data(response)
            chart = response.data['charts'][0]
            self.assertEqual(chart['traces'][0][1][-1], 2)
            self.assertEqual(chart['summary']['wifi_clients'], 2)
            _test_chart_properties(chart)

        self.client.force_login(org2_administrator)
        with self.subTest(
            'Test org admin retrieves metrics for their managed organization'
        ):
            response = self.client.get(path)
            self.assertEqual(response.status_code, 200)
            self._test_response_data(response)
            chart = response.data['charts'][0]
            self.assertEqual(chart['traces'][0][1][-1], 2)
            self.assertEqual(chart['summary']['wifi_clients'], 2)
            _test_chart_properties(chart)

        with self.subTest(
            'Test org admin retrieves metrics for one managed organizations'
        ):
            response = self.client.get(path, {'organization_slug': org2.slug})
            self.assertEqual(response.status_code, 200)
            self._test_response_data(response)
            chart = response.data['charts'][0]
            self.assertEqual(chart['traces'][0][1][-1], 1)
            self.assertEqual(chart['summary']['wifi_clients'], 1)
            _test_chart_properties(chart)

        with self.subTest(
            'Test org admin retrieves metrics for multiple managed organizations'
        ):
            response = self.client.get(
                path, {'organization_slug': f'{org2.slug},{org3.slug}'}
            )
            self.assertEqual(response.status_code, 200)
            self._test_response_data(response)
            chart = response.data['charts'][0]
            self.assertEqual(chart['traces'][0][1][-1], 2)
            self.assertEqual(chart['summary']['wifi_clients'], 2)
            _test_chart_properties(chart)

        with self.subTest('Test org admin retrieve metrics for unmanaged organization'):
            response = self.client.get(path, {'organization_slug': f'{org1.slug}'})
            self.assertEqual(response.status_code, 403)

        with self.subTest(
            'Test org admin retrieve metrics for multiple organizations (unmanaged org included)'
        ):
            response = self.client.get(
                path, {'organization_slug': f'{org1.slug},{org2.slug}'}
            )
            self.assertEqual(response.status_code, 403)

        with self.subTest('Test filtering for non-existing organization'):
            response = self.client.get(path, {'organization_slug': 'non-existing-org'})
            self.assertEqual(response.status_code, 404)

        self.client.force_login(operator)
        with self.subTest('Test non-org admin retrieve metric'):
            response = self.client.get(path)
            self.assertEqual(response.status_code, 403)

    @patch.dict(
        DEFAULT_DASHBOARD_TRAFFIC_CHART,
        {'__all__': ['wan'], org2_id: ['eth1'], org3_id: ['wan', 'eth1']},
    )
    def test_traffic_chart(self):
        def _test_chart_properties(chart):
            self.assertEqual(chart['title'], 'General Traffic')
            self.assertEqual(chart['type'], 'stackedbar+lines')
            self.assertEqual(chart['unit'], 'adaptive_prefix+B')
            self.assertEqual(
                chart['summary_labels'],
                ['Total traffic', 'Total download traffic', 'Total upload traffic'],
            )
            self.assertEqual(chart['colors'], ['#7f7f7f', '#1f77b4', '#ff7f0e'])
            self.assertEqual(chart['colorscale'], None)
            self.assertEqual(
                chart['description'],
                'Network traffic of the whole network (total, download, upload).',
            )

        path = reverse('monitoring_general:api_dashboard_timeseries')
        org1 = self._create_org(name='org1', slug='org1')
        org2 = self._create_org(name='org2', slug='org2', id=self.org2_id)
        org3 = self._create_org(name='org3', slug='org3', id=self.org3_id)
        org1_wan_metric = self._create_org_traffic_metric(org1, 'wan')
        org2_wan_metric = self._create_org_traffic_metric(org2, 'wan')
        org2_eth1_metric = self._create_org_traffic_metric(org2, 'eth1')
        org3_wan_metric = self._create_org_traffic_metric(org3, 'wan')
        org3_eth1_metric = self._create_org_traffic_metric(org3, 'eth1')
        traffic_metrics = {
            'org1': {'wan': 10000000000},
            'org2': {'wan': 20000000000, 'eth1': 60000000000},
            'org3': {'wan': 30000000000, 'eth1': 70000000000},
        }

        org1_wan_metric.write(traffic_metrics['org1']['wan'])
        org2_wan_metric.write(traffic_metrics['org2']['wan'])
        org2_eth1_metric.write(traffic_metrics['org2']['eth1'])
        org3_wan_metric.write(traffic_metrics['org3']['wan'])
        org3_eth1_metric.write(traffic_metrics['org3']['eth1'])

        admin, org2_administrator, operator = self._create_test_users(org2)
        OrganizationUser.objects.create(
            user=org2_administrator, organization=org3, is_admin=True
        )

        self.client.force_login(admin)
        with self.subTest('Test superuser retrieves metric for all organizations'):
            with self.assertNumQueries(3):
                response = self.client.get(path)
            self.assertEqual(response.status_code, 200)
            self._test_response_data(response)
            chart = response.data['charts'][0]
            expected_traffic_data = (
                traffic_metrics['org1']['wan']
                + traffic_metrics['org2']['wan']
                + traffic_metrics['org3']['wan']
            )
            self.assertEqual(
                chart['traces'][0][1][-1], expected_traffic_data / 1000000000
            )
            self.assertEqual(
                chart['summary']['download'], expected_traffic_data / 1000000000
            )
            _test_chart_properties(chart)

        with self.subTest(
            'Test superuser retrieves metric for organization without interface config'
        ):
            response = self.client.get(path, {'organization_slug': org1.slug})
            self.assertEqual(response.status_code, 200)
            self._test_response_data(response)
            chart = response.data['charts'][0]
            self.assertEqual(
                chart['traces'][0][1][-1], traffic_metrics['org1']['wan'] / 1000000000
            )
            self.assertEqual(
                chart['summary']['download'],
                traffic_metrics['org1']['wan'] / 1000000000,
            )
            _test_chart_properties(chart)

        with self.subTest(
            'Test superuser retrieves metric for organization with one interface in config'
        ):
            response = self.client.get(path, {'organization_slug': org2.slug})
            self.assertEqual(response.status_code, 200)
            self._test_response_data(response)
            chart = response.data['charts'][0]
            self.assertEqual(
                chart['traces'][0][1][-1], traffic_metrics['org2']['eth1'] / 1000000000
            )
            self.assertEqual(
                chart['summary']['download'],
                traffic_metrics['org2']['eth1'] / 1000000000,
            )
            _test_chart_properties(chart)

        with self.subTest(
            'Test superuser retrieves metric for organization with multiple interface in config'
        ):
            response = self.client.get(path, {'organization_slug': org3.slug})
            self.assertEqual(response.status_code, 200)
            self._test_response_data(response)
            chart = response.data['charts'][0]
            expected_traffic_data = (
                traffic_metrics['org3']['wan'] + traffic_metrics['org3']['eth1']
            )
            self.assertEqual(
                chart['traces'][0][1][-1], expected_traffic_data / 1000000000
            )
            self.assertEqual(
                chart['summary']['download'], expected_traffic_data / 1000000000
            )
            _test_chart_properties(chart)

        with self.subTest('Test superuser retrieves metric for multiple organization'):
            response = self.client.get(path)
            self.assertEqual(response.status_code, 200)
            self._test_response_data(response)
            chart = response.data['charts'][0]
            expected_traffic_data = (
                traffic_metrics['org1']['wan']
                + traffic_metrics['org2']['wan']
                + traffic_metrics['org3']['wan']
            )
            self.assertEqual(
                chart['traces'][0][1][-1], expected_traffic_data / 1000000000
            )
            self.assertEqual(
                chart['summary']['download'], expected_traffic_data / 1000000000
            )
            _test_chart_properties(chart)

        self.client.force_login(org2_administrator)
        with self.subTest(
            'Test org admin retrieves metrics for their managed organization'
        ):
            response = self.client.get(path)
            self.assertEqual(response.status_code, 200)
            self._test_response_data(response)
            chart = response.data['charts'][0]
            expected_traffic_data = (
                traffic_metrics['org2']['wan'] + traffic_metrics['org3']['wan']
            )
            self.assertEqual(
                chart['traces'][0][1][-1], expected_traffic_data / 1000000000
            )
            self.assertEqual(
                chart['summary']['download'], expected_traffic_data / 1000000000
            )
            _test_chart_properties(chart)

        with self.subTest(
            'Test org admin retrieves metrics for one managed organizations'
        ):
            response = self.client.get(path, {'organization_slug': org2.slug})
            self.assertEqual(response.status_code, 200)
            self._test_response_data(response)
            chart = response.data['charts'][0]
            self.assertEqual(
                chart['traces'][0][1][-1], traffic_metrics['org2']['eth1'] / 1000000000
            )
            self.assertEqual(
                chart['summary']['download'],
                traffic_metrics['org2']['eth1'] / 1000000000,
            )
            _test_chart_properties(chart)

        with self.subTest(
            'Test org admin retrieves metrics for multiple managed organizations'
        ):
            response = self.client.get(
                path, {'organization_slug': f'{org2.slug},{org3.slug}'}
            )
            self.assertEqual(response.status_code, 200)
            self._test_response_data(response)
            chart = response.data['charts'][0]
            expected_traffic_data = (
                traffic_metrics['org2']['wan'] + traffic_metrics['org3']['wan']
            )
            self.assertEqual(
                chart['traces'][0][1][-1], expected_traffic_data / 1000000000
            )
            self.assertEqual(
                chart['summary']['download'], expected_traffic_data / 1000000000
            )
            _test_chart_properties(chart)

        with self.subTest('Test org admin retrieve metrics for unmanaged organization'):
            response = self.client.get(path, {'organization_slug': f'{org1.slug}'})
            self.assertEqual(response.status_code, 403)

        with self.subTest(
            'Test org admin retrieve metrics for multiple organizations (unmanaged org included)'
        ):
            response = self.client.get(
                path, {'organization_slug': f'{org1.slug},{org2.slug}'}
            )
            self.assertEqual(response.status_code, 403)

        # org2_administrator will only have one managed organization, i.e. org3
        org2.delete()
        with self.subTest(
            'Test org admin retrieves metric for only managed organization'
        ):
            response = self.client.get(path)
            self.assertEqual(response.status_code, 200)
            self._test_response_data(response)
            chart = response.data['charts'][0]
            expected_traffic_data = (
                traffic_metrics['org3']['wan'] + traffic_metrics['org3']['eth1']
            )
            self.assertEqual(
                chart['traces'][0][1][-1], expected_traffic_data / 1000000000
            )
            self.assertEqual(
                chart['summary']['download'], expected_traffic_data / 1000000000
            )
            _test_chart_properties(chart)

        with self.subTest('Test filtering for non-existing organization'):
            response = self.client.get(path, {'organization_slug': 'non-existing-org'})
            self.assertEqual(response.status_code, 404)

        self.client.force_login(operator)
        with self.subTest('Test non-org admin retrieve metric'):
            response = self.client.get(path)
            self.assertEqual(response.status_code, 403)

        with self.subTest('Test view cache'):
            with self.assertNumQueries(1):
                response = self.client.get(path)

    def test_exporting_data_as_csv(self):
        def _test_csv_response(response):
            self.assertEqual(
                response.get('Content-Disposition'), 'attachment; filename=data.csv'
            )
            self.assertEqual(response.get('Content-Type'), 'text/csv')
            rows = response.content.decode('utf8').strip().split('\n')
            header = rows[0].strip().split(',')
            self.assertEqual(
                header,
                [
                    'time',
                    'wifi_clients - General WiFi Clients',
                    'download - General Traffic',
                    'upload - General Traffic',
                ],
            )
            last_line = rows[-1].strip().split(',')
            return last_line

        path = reverse('monitoring_general:api_dashboard_timeseries')
        org1 = self._create_org(name='org1', slug='org1')
        org2 = self._create_org(name='org2', slug='org2')
        org1_wifi_metric = self._create_org_wifi_client_metric(org1)
        org2_wifi_metric = self._create_org_wifi_client_metric(org2)
        org1_wan_metric = self._create_org_traffic_metric(org1, 'wan')
        org2_wan_metric = self._create_org_traffic_metric(org2, 'wan')

        org1_wifi_metric.write('00:23:4a:00:00:00')
        org2_wifi_metric.write('00:23:4b:00:00:00')
        org2_wifi_metric.write('00:14:5c:00:00:00')
        org1_wan_metric.write(1000000000)
        org2_wan_metric.write(2000000000)

        admin, org2_administrator, operator = self._create_test_users(org2)

        self.client.force_login(admin)
        with self.subTest('Test superuser exporting csv'):
            response = self.client.get(path, {'csv': '1'})
            last_line = _test_csv_response(response)
            self.assertEqual(
                last_line,
                [last_line[0], '3', '3.0', ''],
            )

        self.client.force_login(org2_administrator)
        with self.subTest('Test organization admin exporting csv'):
            response = self.client.get(path, {'csv': '1'})
            last_line = _test_csv_response(response)
            self.assertEqual(
                last_line,
                [last_line[0], '2', '2.0', ''],
            )

        self.client.force_login(operator)
        with self.subTest('Test non-organization admin exporting csv'):
            response = self.client.get(path, {'csv': 1})
            self.assertEqual(response.status_code, 403)

        self.client.logout()
        with self.subTest('Test unauthenticated user exporting csv'):
            response = self.client.get(path, {'csv': 1})
            self.assertEqual(response.status_code, 401)

    def test_filter_with_location_id(self):
        path = reverse('monitoring_general:api_dashboard_timeseries')
        org1 = self._create_org(name='org1', slug='org1')
        org2 = self._create_org(name='org2', slug='org2')
        org3 = self._create_org(name='org3', slug='org3')
        org1_location = self._create_location(organization=org1)
        org2_location = self._create_location(organization=org2)
        org3_location = self._create_location(organization=org3)
        org1_metric = self._create_org_wifi_client_metric(
            org1, location_id=str(org1_location.id)
        )
        org2_metric = self._create_org_wifi_client_metric(
            org2, location_id=str(org2_location.id)
        )
        org3_metric = self._create_org_wifi_client_metric(
            org3, location_id=str(org3_location.id)
        )

        org1_metric.write('00:23:4a:00:00:00')
        org2_metric.write('00:23:4b:00:00:00')
        org3_metric.write('00:14:5c:00:00:00')

        admin, org2_administrator, operator = self._create_test_users(org2)
        OrganizationUser.objects.create(
            user=org2_administrator, organization=org3, is_admin=True
        )

        self.client.force_login(admin)
        with self.subTest('Test superuser filter with one location_id'):
            response = self.client.get(path, {'location_id': str(org1_location.id)})
            self.assertEqual(response.status_code, 200)
            chart = response.data['charts'][0]
            self.assertEqual(chart['traces'][0][1][-1], 1)
            self.assertEqual(chart['summary']['wifi_clients'], 1)

        with self.subTest('Test superuser filter with multiple location_id'):
            response = self.client.get(
                path, {'location_id': f'{org1_location.id},{org2_location.id}'}
            )
            self.assertEqual(response.status_code, 200)
            chart = response.data['charts'][0]
            self.assertEqual(chart['traces'][0][1][-1], 2)
            self.assertEqual(chart['summary']['wifi_clients'], 2)

        self.client.force_login(org2_administrator)
        with self.subTest('Test organization admin filter with one location_id'):
            response = self.client.get(path, {'location_id': str(org2_location.id)})
            self.assertEqual(response.status_code, 200)
            chart = response.data['charts'][0]
            self.assertEqual(chart['traces'][0][1][-1], 1)
            self.assertEqual(chart['summary']['wifi_clients'], 1)

        with self.subTest('Test organization admin filter with multiple location_ids'):
            response = self.client.get(
                path, {'location_id': f'{org2_location.id},{org3_location.id}'}
            )
            self.assertEqual(response.status_code, 200)
            chart = response.data['charts'][0]
            self.assertEqual(chart['traces'][0][1][-1], 2)
            self.assertEqual(chart['summary']['wifi_clients'], 2)

        with self.subTest(
            'Test organization admin filter location_id of another organization'
        ):
            response = self.client.get(path, {'location_id': str(org1_location.id)})
            self.assertEqual(response.status_code, 404)

        with self.subTest(
            'Test organization admin filter with multiple location_ids (includes other org)'
        ):
            response = self.client.get(
                path, {'location_id': f'{org1_location.id},{org2_location.id}'}
            )
            self.assertEqual(response.status_code, 404)

        self.client.force_login(operator)
        with self.subTest('Test non-organization admin filter location_id'):
            response = self.client.get(path, {'location_id': str(org1_location.id)})
            self.assertEqual(response.status_code, 403)

        with self.subTest('Test unauthenticated user filter location_id'):
            response = self.client.get(path, {'location_id': str(org1_location.id)})
            self.assertEqual(response.status_code, 403)

    def test_filter_with_floorplan_id(self):
        path = reverse('monitoring_general:api_dashboard_timeseries')
        org1 = self._create_org(name='org1', slug='org1')
        org2 = self._create_org(name='org2', slug='org2')
        org3 = self._create_org(name='org3', slug='org3')
        org1_location = self._create_location(organization=org1, type='indoor')
        org1_floorplan = self._create_floorplan(location=org1_location)
        org2_location = self._create_location(organization=org2, type='indoor')
        org2_floorplan = self._create_floorplan(location=org2_location)
        org3_location = self._create_location(organization=org3, type='indoor')
        org3_floorplan = self._create_floorplan(location=org3_location)
        org1_metric = self._create_org_wifi_client_metric(
            org1, floorplan_id=str(org1_floorplan.id), location=str(org1_floorplan.id)
        )
        org2_metric = self._create_org_wifi_client_metric(
            org2, floorplan_id=str(org2_floorplan.id), location=str(org2_floorplan.id)
        )
        org3_metric = self._create_org_wifi_client_metric(
            org3, floorplan_id=str(org3_floorplan.id), location=str(org3_floorplan.id)
        )

        org1_metric.write('00:23:4a:00:00:00')
        org2_metric.write('00:23:4b:00:00:00')
        org3_metric.write('00:14:5c:00:00:00')

        admin, org2_administrator, operator = self._create_test_users(org2)
        OrganizationUser.objects.create(
            user=org2_administrator, organization=org3, is_admin=True
        )

        self.client.force_login(admin)
        with self.subTest('Test superuser filter with one floorplan_id'):
            response = self.client.get(path, {'floorplan_id': str(org1_floorplan.id)})
            self.assertEqual(response.status_code, 200)
            chart = response.data['charts'][0]
            self.assertEqual(chart['traces'][0][1][-1], 1)
            self.assertEqual(chart['summary']['wifi_clients'], 1)

        with self.subTest('Test superuser filter with multiple floorplan_id'):
            response = self.client.get(
                path, {'floorplan_id': f'{org1_floorplan.id},{org2_floorplan.id}'}
            )
            self.assertEqual(response.status_code, 200)
            chart = response.data['charts'][0]
            self.assertEqual(chart['traces'][0][1][-1], 2)
            self.assertEqual(chart['summary']['wifi_clients'], 2)

        self.client.force_login(org2_administrator)
        with self.subTest('Test organization admin filter with one floorplan_id'):
            response = self.client.get(path, {'floorplan_id': str(org2_floorplan.id)})
            self.assertEqual(response.status_code, 200)
            chart = response.data['charts'][0]
            self.assertEqual(chart['traces'][0][1][-1], 1)
            self.assertEqual(chart['summary']['wifi_clients'], 1)

        with self.subTest('Test organization admin filter with multiple floorplan_ids'):
            response = self.client.get(
                path, {'floorplan_id': f'{org2_floorplan.id},{org3_floorplan.id}'}
            )
            self.assertEqual(response.status_code, 200)
            chart = response.data['charts'][0]
            self.assertEqual(chart['traces'][0][1][-1], 2)
            self.assertEqual(chart['summary']['wifi_clients'], 2)

        with self.subTest(
            'Test organization admin filter floorplan_id of another organization'
        ):
            response = self.client.get(path, {'floorplan_id': str(org1_floorplan.id)})
            self.assertEqual(response.status_code, 404)

        with self.subTest(
            'Test organization admin filter with multiple floorplan_ids (includes other org)'
        ):
            response = self.client.get(
                path, {'floorplan_id': f'{org1_floorplan.id},{org2_floorplan.id}'}
            )
            self.assertEqual(response.status_code, 404)

        self.client.force_login(operator)
        with self.subTest('Test non-organization admin filter floorplan_id'):
            response = self.client.get(path, {'floorplan_id': str(org1_floorplan.id)})
            self.assertEqual(response.status_code, 403)

        with self.subTest('Test unauthenticated user filter floorplan_id'):
            response = self.client.get(path, {'floorplan_id': str(org1_floorplan.id)})
            self.assertEqual(response.status_code, 403)

    def test_group_by_time(self):
        admin = self._create_admin()
        self.client.force_login(admin)
        path = reverse('monitoring_general:api_dashboard_timeseries')
        with self.subTest('Test with valid group time'):
            response = self.client.get(path, {'time': '1d'})
            self.assertEqual(response.status_code, 200)
            self.assertIsInstance(response.data['charts'], list)

        with self.subTest('Test with custom valid group time'):
            response = self.client.get(path, {'time': '5d'})
            self.assertEqual(response.status_code, 200)
            self.assertIsInstance(response.data['charts'], list)

        with self.subTest('Test with invalid group time'):
            response = self.client.get(path, {'time': '3w'})
            self.assertEqual(response.status_code, 400)

    def test_organizations_list(self):
        path = reverse('monitoring_general:api_dashboard_timeseries')
        Organization.objects.all().delete()
        org1 = self._create_org(name='org1', slug='org1')
        admin, org_admin, _ = self._create_test_users(org1)

        self.client.force_login(admin)
        with self.subTest('Superuser: Only one organization is present'):
            response = self.client.get(path)
            self.assertEqual(response.status_code, 200)
            self.assertNotIn('organizations', response.data)

        org2 = self._create_org(name='org2', slug='org2', id=self.org2_id)

        with self.subTest('Superuser: Multiple organizations are present'):
            response = self.client.get(path)
            self.assertEqual(response.status_code, 200)
            self.assertIn('organizations', response.data)
            self.assertEqual(len(response.data['organizations']), 2)
            self.assertEqual(
                response.data['organizations'],
                [{'id': org.slug, 'text': org.name} for org in [org1, org2]],
            )

        self.client.logout()
        self.client.force_login(org_admin)

        with self.subTest('Non-superuser: Administrator of one organization'):
            response = self.client.get(path)
            self.assertEqual(response.status_code, 200)
            self.assertNotIn('organizations', response.data)

        OrganizationUser.objects.create(
            user=org_admin, organization=org2, is_admin=True
        )

        with self.subTest('Non-superuser: Administrator of multiple organizations'):
            response = self.client.get(path)
            self.assertEqual(response.status_code, 200)
            self.assertIn('organizations', response.data)
            self.assertEqual(len(response.data['organizations']), 2)
            self.assertEqual(
                response.data['organizations'],
                [{'id': org.slug, 'text': org.name} for org in [org1, org2]],
            )



# NetJSON DeviceMonitoring schema,
# https://github.com/netjson/netjson/blob/master/schema/device-monitoring.json
schema = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "$id": "https://raw.githubusercontent.com/netjson/netjson/master/schema/device-monitoring.json",
    "title": "NetJSON Device Monitoring",
    "description": "Monitoring information sent by a device.",
    "type": "object",
    "additionalProperties": True,
    "required": ["type"],
    "properties": {
        "type": {"type": "string", "enum": ["DeviceMonitoring"]},
        "general": {
            "type": "object",
            "title": "General",
            "additionalProperties": True,
            "properties": {
                "local_time": {"type": "integer"},
                "uptime": {"type": "integer"},
            },
        },
        "resources": {
            "type": "object",
            "title": "Resources",
            "additionalProperties": True,
            "properties": {
                "load": {
                    "type": "array",
                    "items": {"type": "number", "minItems": 3, "maxItems": 3},
                },
                "cpus": {"type": "integer"},
                "memory": {
                    "id": "memory",
                    "type": "object",
                    "properties": {
                        "total": {"type": "integer"},
                        "free": {"type": "integer"},
                        "buffered": {"type": "integer"},
                        "cached": {"type": "integer"},
                        "shared": {"type": "integer"},
                        "available": {"type": "integer"},
                    },
                    "required": ["total", "free", "buffered", "shared"],
                },
                "disk": {
                    "type": "array",
                    "additionalItems": False,
                    "title": "Disks",
                    "items": {
                        "type": "object",
                        "title": "Disk",
                        "additionalProperties": False,
                        "properties": {
                            "mount_point": {"type": "string"},
                            "filesystem": {"type": "string"},
                            "used_bytes": {"type": "integer"},
                            "available_bytes": {"type": "integer"},
                            "used_percent": {"type": "integer"},
                            "size_bytes": {"type": "integer"},
                        },
                        "required": [
                            "mount_point",
                            "filesystem",
                            "used_bytes",
                            "available_bytes",
                            "used_percent",
                            "size_bytes",
                        ],
                    },
                },
                "swap": {
                    "type": "object",
                    "properties": {
                        "total": {"type": "integer"},
                        "free": {"type": "integer"},
                    },
                },
                "connections": {
                    "type": "object",
                    "properties": {
                        "ipv4": {
                            "type": "object",
                            "properties": {
                                "tcp": {"type": "integer"},
                                "udp": {"type": "integer"},
                            },
                        },
                        "ipv6": {
                            "type": "object",
                            "properties": {
                                "tcp": {"type": "integer"},
                                "udp": {"type": "integer"},
                            },
                        },
                    },
                },
                "processes": {
                    "type": "object",
                    "properties": {
                        "running": {"type": "integer"},
                        "sleeping": {"type": "integer"},
                        "blocked": {"type": "integer"},
                        "zombie": {"type": "integer"},
                        "stopped": {"type": "integer"},
                        "paging": {"type": "integer"},
                    },
                },
                "cpu": {
                    "type": "object",
                    "properties": {
                        "frequency": {"type": "integer"},
                        "user": {"type": "integer"},
                        "system": {"type": "integer"},
                        "nice": {"type": "integer"},
                        "idle": {"type": "integer"},
                        "iowait": {"type": "integer"},
                        "irq": {"type": "integer"},
                        "softirq": {"type": "integer"},
                    },
                },
                "flash": {
                    "type": "object",
                    "properties": {
                        "total": {"type": "integer"},
                        "free": {"type": "integer"},
                    },
                },
                "storage": {
                    "type": "object",
                    "properties": {
                        "total": {"type": "integer"},
                        "free": {"type": "integer"},
                    },
                },
            },
        },
        "interfaces": {
            "type": "array",
            "title": "Interfaces",
            "uniqueItems": True,
            "additionalItems": False,
            "items": {
                "type": "object",
                "title": "Interface",
                "additionalProperties": True,
                "required": ["name"],
                "properties": {
                    "name": {"type": "string"},
                    "type": {"type": "string"},
                    "uptime": {"type": "integer"},
                    "up": {"type": "boolean"},
                    "mac": {"type": "string"},
                    "mtu": {"type": "integer"},
                    "txqueuelen": {"type": "integer"},
                    "speed": {"type": "string"},
                    "multicast": {"type": "boolean"},
                    "bridge_members": {
                        "type": "array",
                        "items": {"type": "string"},
                        "additionalItems": False,
                    },
                    "stp": {"type": "boolean"},
                    "statistics": {
                        "type": "object",
                        "properties": {
                            "collisions": {"type": "integer"},
                            "rx_frame_errors": {"type": "integer"},
                            "tx_compressed": {"type": "integer"},
                            "multicast": {"type": "integer"},
                            "rx_length_errors": {"type": "integer"},
                            "tx_dropped": {"type": "integer"},
                            "rx_bytes": {"type": "integer"},
                            "rx_missed_errors": {"type": "integer"},
                            "tx_errors": {"type": "integer"},
                            "rx_compressed": {"type": "integer"},
                            "rx_over_errors": {"type": "integer"},
                            "tx_fifo_errors": {"type": "integer"},
                            "rx_crc_errors": {"type": "integer"},
                            "rx_packets": {"type": "integer"},
                            "tx_heartbeat_errors": {"type": "integer"},
                            "rx_dropped": {"type": "integer"},
                            "tx_aborted_errors": {"type": "integer"},
                            "tx_packets": {"type": "integer"},
                            "rx_errors": {"type": "integer"},
                            "tx_bytes": {"type": "integer"},
                            "tx_window_errors": {"type": "integer"},
                            "rx_fifo_errors": {"type": "integer"},
                            "tx_carrier_errors": {"type": "integer"},
                        },
                    },
                    "wireless": {
                        "type": "object",
                        "additionalProperties": True,
                        "required": [
                            "channel",
                            "country",
                            "frequency",
                            "mode",
                            "noise",
                            "ssid",
                            "tx_power",
                        ],
                        "properties": {
                            "channel": {"type": "integer"},
                            "country": {"type": "string"},
                            "frequency": {"type": "integer"},
                            "mode": {"type": "string"},
                            "noise": {"type": "integer"},
                            "signal": {"type": "integer"},
                            "ssid": {"type": "string"},
                            "tx_power": {"type": "integer"},
                            "htmode": {"type": "string"},
                            "clients": {
                                "type": "array",
                                "title": "Wireless Clients",
                                "additionalItems": False,
                                "items": {
                                    "type": "object",
                                    "title": "Client",
                                    "additionalProperties": True,
                                    "required": [
                                        "auth",
                                        "authorized",
                                        "ht",
                                        "vht",
                                        "wmm",
                                        "mac",
                                        "mfp",
                                    ],
                                    "properties": {
                                        "aid": {"type": "integer"},
                                        "signal": {"type": "integer"},
                                        "noise": {"type": "integer"},
                                        "assoc": {"type": "boolean"},
                                        "auth": {"type": "boolean"},
                                        "authorized": {"type": "boolean"},
                                        "ht": {"type": "boolean"},
                                        "vht": {"type": "boolean"},
                                        "he": {"type": "boolean"},
                                        "wds": {"type": "boolean"},
                                        "wmm": {"type": "boolean"},
                                        "wps": {"type": "boolean"},
                                        "mac": {"type": "string"},
                                        "bitrate": {"type": "integer"},
                                        "quality": {"type": "integer"},
                                        "quality_max": {"type": "integer"},
                                        "vendor": {"type": "string"},
                                        "mfp": {"type": "boolean"},
                                        "preauth": {"type": "boolean"},
                                        "signature": {"type": "string"},
                                        "rrm": {
                                            "type": "array",
                                            "minItems": 5,
                                            "maxItems": 5,
                                            "items": {"type": "integer"},
                                        },
                                    },
                                },
                            },
                        },
                    },
                    "mobile": {
                        "type": "object",
                        "additionalProperties": True,
                        "required": [
                            "imei",
                            "operator_code",
                            "operator_name",
                            "connection_status",
                            "power_status",
                            "manufacturer",
                            "model",
                        ],
                        "properties": {
                            "imei": {
                                "type": "string",
                                "minLength": 15,
                                "maxLength": 17,
                                "pattern": "^[0-9]*$",
                            },
                            "operator_name": {"type": "string"},
                            "operator_code": {
                                "type": "string",
                                "minLength": 4,
                                "maxLength": 6,
                                "pattern": "^[0-9]*$",
                            },
                            "connection_status": {"type": "string"},
                            "power_status": {"type": "string"},
                            "manufacturer": {"type": "string"},
                            "model": {"type": "string"},
                            "signal": {
                                "type": "object",
                                "additionalProperties": True,
                                "properties": {
                                    "gsm": {
                                        "type": "object",
                                        "additionalProperties": True,
                                        "required": ["rssi"],
                                        "properties": {"rssi": {"type": "number"}},
                                    },
                                    "cdma1x": {
                                        "type": "object",
                                        "additionalProperties": True,
                                        "required": ["rssi", "ecio"],
                                        "properties": {
                                            "rssi": {"type": "number"},
                                            "ecio": {"type": "number"},
                                        },
                                    },
                                    "evdo": {
                                        "type": "object",
                                        "additionalProperties": True,
                                        "required": ["rssi", "ecio", "io", "sinr"],
                                        "properties": {
                                            "rssi": {"type": "number"},
                                            "ecio": {"type": "number"},
                                            "io": {"type": "number"},
                                            "sinr": {"type": "number"},
                                        },
                                    },
                                    "umts": {
                                        "type": "object",
                                        "additionalProperties": True,
                                        "required": ["ecio"],
                                        "properties": {
                                            "rssi": {"type": "number"},
                                            "rscp": {"type": "number"},
                                            "ecio": {"type": "number"},
                                        },
                                    },
                                    "lte": {
                                        "type": "object",
                                        "additionalProperties": True,
                                        "required": ["rssi", "rsrp", "rsrq", "snr"],
                                        "properties": {
                                            "rssi": {"type": "number"},
                                            "rsrp": {"type": "number"},
                                            "rsrq": {"type": "number"},
                                            "snr": {"type": "number"},
                                        },
                                    },
                                    "5g": {
                                        "type": "object",
                                        "additionalProperties": True,
                                        "required": ["rsrp", "rsrq", "snr"],
                                        "properties": {
                                            "rsrp": {"type": "number"},
                                            "rsrq": {"type": "number"},
                                            "snr": {"type": "number"},
                                        },
                                    },
                                },
                            },
                        },
                    },
                    "addresses": {
                        "type": "array",
                        "title": "Addresses",
                        "uniqueItems": True,
                        "additionalItems": False,
                        "items": {
                            "additionalProperties": True,
                            "title": "Address",
                            "type": "object",
                            "required": ["proto", "family", "address", "mask"],
                            "properties": {
                                "proto": {"type": "string"},
                                "family": {"type": "string"},
                                "address": {
                                    "type": "string",
                                    "anyOf": [{"format": "ipv4"}, {"format": "ipv6"}],
                                },
                                "mask": {"type": "integer"},
                            },
                        },
                    },
                },
            },
        },
        "dhcp_leases": {
            "type": "array",
            "title": "DHCP leases",
            "additionalItems": False,
            "items": {
                "type": "object",
                "title": "DHCP lease",
                "additionalProperties": False,
                "properties": {
                    "expiry": {"type": "number"},
                    "ip": {
                        "type": "string",
                        "anyOf": [{"format": "ipv4"}, {"format": "ipv6"}],
                    },
                    "mac": {"type": "string"},
                    "vendor": {"type": "string"},
                    "client_name": {"type": "string"},
                    "client_id": {"type": "string"},
                },
                "required": ["expiry", "ip", "mac", "client_name", "client_id"],
            },
        },
        "neighbors": {
            "type": "array",
            "title": "Neighbors",
            "additionalItems": False,
            "items": {
                "type": "object",
                "title": "Neighbor",
                "additionalProperties": False,
                "properties": {
                    "ip": {
                        "type": "string",
                        "anyOf": [{"format": "ipv4"}, {"format": "ipv6"}],
                    },
                    "mac": {"type": "string"},
                    "vendor": {"type": "string"},
                    "interface": {"type": "string"},
                    "state": {"type": "string"},
                },
                "required": ["ip", "interface"],
            },
        },
    },
}



from django.contrib import admin
from django.forms import ModelForm
from django.utils.translation import gettext_lazy as _
from reversion.admin import VersionAdmin
from swapper import load_model

from openwisp_utils.admin import TimeReadonlyAdminMixin

Chart = load_model('monitoring', 'Chart')
Metric = load_model('monitoring', 'Metric')
AlertSettings = load_model('monitoring', 'AlertSettings')
Device = load_model('config', 'Device')
Organization = load_model('openwisp_users', 'Organization')


class AlertSettingsForm(ModelForm):
    def __init__(self, *args, **kwargs):
        instance = kwargs.get('instance')
        if instance:
            kwargs['initial'] = {
                'custom_tolerance': instance.tolerance,
                'custom_threshold': instance.threshold,
                'custom_operator': instance.operator,
            }
        super().__init__(*args, **kwargs)


class AlertSettingsInline(TimeReadonlyAdminMixin, admin.StackedInline):
    model = AlertSettings
    form = AlertSettingsForm
    extra = 0


class ChartInline(admin.StackedInline):
    model = Chart
    extra = 0
    template = 'admin/chart_inline.html'
    exclude = ['created', 'modified']


@admin.register(Metric)
class MetricAdmin(TimeReadonlyAdminMixin, VersionAdmin):
    list_display = ['__str__', 'created', 'modified']
    readonly_fields = ['is_healthy']
    search_fields = ['name']
    save_on_top = True
    inlines = [ChartInline, AlertSettingsInline]
    fieldsets = [
        (None, {'fields': ('name', 'content_type', 'object_id', 'configuration')}),
        (
            _('Advanced options'),
            {'classes': ('collapse',), 'fields': ('key', 'field_name')},
        ),
    ]

    class Media:
        css = {'all': ('monitoring/css/monitoring.css', 'monitoring/css/chart.css')}
        js = ('monitoring/js/lib/plotly-cartesian.min.js', 'monitoring/js/chart.js')

    def reversion_register(self, model, **options):
        if model == Metric:
            options['follow'] = (
                *(options['follow']),
                'content_object',
                'chart_set',
            )
        if model == AlertSettings:
            options['follow'] = (
                *(options['follow']),
                'metric',
            )
        if model == Chart:
            options['follow'] = (*options['follow'], 'metric')
        return super().reversion_register(model, **options)



# Manually created

import logging

from django.db import migrations

from .influxdb.influxdb_alter_structure_0006 import (
    update_metric_timeseries_structure_forward_migration,
    update_metric_timeseries_structure_reverse_migration,
)

CHUNK_SIZE = 1000

logger = logging.getLogger(__name__)


def forward_migration(apps, schema_editor):
    update_metric_timeseries_structure_forward_migration(apps, schema_editor)
    from ..tasks import migrate_timeseries_database

    migrate_timeseries_database.delay()


def reverse_migration(apps, schema_editor):
    update_metric_timeseries_structure_reverse_migration(
        apps, schema_editor, metric_keys=['signal']
    )


class Migration(migrations.Migration):
    dependencies = [('monitoring', '0011_alter_metric_field_name')]

    operations = [
        migrations.RunPython(
            forward_migration,
            reverse_code=reverse_migration,
        )
    ]



# Generated by Django 3.1.2 on 2020-11-16 09:03

import uuid

import django.core.validators
import django.db.migrations.operations.special
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import swapper
from django.conf import settings
from django.db import migrations, models
from swapper import split

from ..configuration import CHART_CONFIGURATION_CHOICES, METRIC_CONFIGURATION_CHOICES


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        swapper.dependency(
            *split(settings.AUTH_USER_MODEL), version='0004_default_groups'
        ),
        ('contenttypes', '0002_remove_content_type_name'),
        swapper.dependency('openwisp_notifications', 'Notifications'),
    ]

    operations = [
        migrations.CreateModel(
            name='Metric',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'created',
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='created',
                    ),
                ),
                (
                    'modified',
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='modified',
                    ),
                ),
                ('name', models.CharField(max_length=64)),
                (
                    'configuration',
                    models.CharField(
                        choices=METRIC_CONFIGURATION_CHOICES, max_length=16, null=True
                    ),
                ),
                (
                    'key',
                    models.SlugField(
                        blank=True,
                        help_text='leave blank to determine automatically',
                        max_length=64,
                    ),
                ),
                ('field_name', models.CharField(default='value', max_length=16)),
                (
                    'object_id',
                    models.CharField(blank=True, db_index=True, max_length=36),
                ),
                (
                    'content_type',
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to='contenttypes.contenttype',
                    ),
                ),
                (
                    'is_healthy',
                    models.BooleanField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
            ],
            options={
                'abstract': False,
                'swappable': swapper.swappable_setting('monitoring', 'Metric'),
                'unique_together': {('key', 'field_name', 'content_type', 'object_id')},
            },
        ),
        migrations.CreateModel(
            name='Chart',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'created',
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='created',
                    ),
                ),
                (
                    'modified',
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='modified',
                    ),
                ),
                (
                    'configuration',
                    models.CharField(
                        choices=CHART_CONFIGURATION_CHOICES, max_length=16, null=True
                    ),
                ),
                (
                    'metric',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name('monitoring', 'Metric'),
                    ),
                ),
            ],
            options={
                'abstract': False,
                'swappable': swapper.swappable_setting('monitoring', 'Chart'),
            },
        ),
        migrations.CreateModel(
            name='AlertSettings',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'created',
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='created',
                    ),
                ),
                (
                    'modified',
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='modified',
                    ),
                ),
                (
                    'custom_operator',
                    models.CharField(
                        blank=True,
                        choices=[('<', 'less than'), ('>', 'greater than')],
                        max_length=1,
                        null=True,
                        verbose_name='operator',
                    ),
                ),
                (
                    'custom_threshold',
                    models.FloatField(
                        blank=True,
                        help_text='threshold value',
                        null=True,
                        verbose_name='threshold value',
                    ),
                ),
                (
                    'custom_tolerance',
                    models.PositiveIntegerField(
                        blank=True,
                        help_text='for how many minutes should the threshold value be crossed before'
                        ' an alert is sent? A value of zero means the alert is sent immediately',
                        null=True,
                        validators=[django.core.validators.MaxValueValidator(10080)],
                        verbose_name='threshold tolerance',
                    ),
                ),
                (
                    'metric',
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name('monitoring', 'Metric'),
                    ),
                ),
                (
                    'is_active',
                    models.BooleanField(
                        default=True,
                        help_text='whether alerts are enabled for this metric, uncheck to disable'
                        ' this alert for this object and all users',
                        verbose_name='Alerts enabled',
                    ),
                ),
            ],
            options={
                'verbose_name': 'Alert settings',
                'verbose_name_plural': 'Alert settings',
                'abstract': False,
                'swappable': swapper.swappable_setting('monitoring', 'AlertSettings'),
            },
        ),
    ]



from unittest.mock import patch

from django.core.exceptions import ImproperlyConfigured

from . import DeviceMonitoringTestCase


class TestSettings(DeviceMonitoringTestCase):
    """Tests ``OpenWISP Device settings`` functionality"""

    @patch(
        'django.conf.settings.OPENWISP_MONITORING_CRITICAL_DEVICE_METRICS',
        [{}],
        create=True,
    )
    def test_invalid_critical_device_metrics_setting(self):
        with self.assertRaises(ImproperlyConfigured):
            from ..settings import get_critical_device_metrics

            get_critical_device_metrics()



# Generated by Django 3.1.2 on 2020-12-20 06:15


from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('device_monitoring', '0001_squashed_0002_devicemonitoring'),
    ]

    operations = [
        # NOTE: Usage of "Monitoring Template" is deprecated in favour of
        # "openwisp-openwisp-monitoring" package. See
        # https://github.com/openwisp/openwrt-openwisp-monitoring#openwrt-openwisp-monitoring
        migrations.RunPython(
            migrations.RunPython.noop, reverse_code=migrations.RunPython.noop
        )
    ]



import logging
import operator
import re
import sys
from collections import OrderedDict
from datetime import datetime

from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.functional import cached_property
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from influxdb import InfluxDBClient
from influxdb.exceptions import InfluxDBClientError
from influxdb.line_protocol import make_lines

from openwisp_monitoring.utils import retry

from ...exceptions import TimeseriesWriteException
from .. import TIMESERIES_DB

logger = logging.getLogger(__name__)


class DatabaseClient(object):
    _AGGREGATE = [
        'COUNT',
        'DISTINCT',
        'INTEGRAL',
        'MEAN',
        'MEDIAN',
        'MODE',
        'SPREAD',
        'STDDEV',
        'SUM',
        'BOTTOM',
        'FIRST',
        'LAST',
        'MAX',
        'MIN',
        'PERCENTILE',
        'SAMPLE',
        'TOP',
        'CEILING',
        'CUMULATIVE_SUM',
        'DERIVATIVE',
        'DIFFERENCE',
        'ELAPSED',
        'FLOOR',
        'HISTOGRAM',
        'MOVING_AVERAGE',
        'NON_NEGATIVE_DERIVATIVE',
        'HOLT_WINTERS',
    ]
    _FORBIDDEN = ['drop', 'create', 'delete', 'alter', 'into']
    backend_name = 'influxdb'

    def __init__(self, db_name=None):
        self._db = None
        self.db_name = db_name or TIMESERIES_DB['NAME']
        self.client_error = InfluxDBClientError

    @retry
    def create_database(self):
        """Creates database if necessary."""
        # InfluxDB does not create a new database, neither raise an error if database exists
        self.db.create_database(self.db_name)
        logger.debug(f'Created InfluxDB database "{self.db_name}"')

    @retry
    def drop_database(self):
        """Drops database if it exists."""
        # InfluxDB does not raise an error if database does not exist
        self.db.drop_database(self.db_name)
        logger.debug(f'Dropped InfluxDB database "{self.db_name}"')

    @cached_property
    def db(self):
        """Returns an ``InfluxDBClient`` instance."""
        return self.dbs['default']

    @cached_property
    def dbs(self):
        dbs = {
            'default': InfluxDBClient(
                TIMESERIES_DB['HOST'],
                TIMESERIES_DB['PORT'],
                TIMESERIES_DB['USER'],
                TIMESERIES_DB['PASSWORD'],
                self.db_name,
                use_udp=TIMESERIES_DB.get('OPTIONS', {}).get('udp_writes', False),
                udp_port=TIMESERIES_DB.get('OPTIONS', {}).get('udp_port', 8089),
            ),
        }
        if TIMESERIES_DB.get('OPTIONS', {}).get('udp_writes', False):
            # When using UDP, InfluxDB allows only using one retention policy
            # per port. Therefore, we need to have different instances of
            # InfluxDBClient.
            dbs['short'] = InfluxDBClient(
                TIMESERIES_DB['HOST'],
                TIMESERIES_DB['PORT'],
                TIMESERIES_DB['USER'],
                TIMESERIES_DB['PASSWORD'],
                self.db_name,
                use_udp=TIMESERIES_DB.get('OPTIONS', {}).get('udp_writes', False),
                udp_port=TIMESERIES_DB.get('OPTIONS', {}).get('udp_port', 8089) + 1,
            )
            dbs['__all__'] = InfluxDBClient(
                TIMESERIES_DB['HOST'],
                TIMESERIES_DB['PORT'],
                TIMESERIES_DB['USER'],
                TIMESERIES_DB['PASSWORD'],
                self.db_name,
            )
        else:
            dbs['short'] = dbs['default']
            dbs['__all__'] = dbs['default']
        return dbs

    @cached_property
    def use_udp(self):
        return TIMESERIES_DB.get('OPTIONS', {}).get('udp_writes', False)

    @retry
    def create_or_alter_retention_policy(self, name, duration):
        """Creates or alters existing retention policy if necessary."""
        retention_policies = self.db.get_list_retention_policies()
        exists = False
        duration_changed = False
        for policy in retention_policies:
            if policy['name'] == name:
                exists = True
                duration_changed = policy['duration']
                break
        if not exists:
            self.db.create_retention_policy(name=name, duration=duration, replication=1)
        elif exists and duration_changed:
            self.db.alter_retention_policy(name=name, duration=duration)

    @retry
    def query(self, query, precision=None, **kwargs):
        database = kwargs.get('database') or self.db_name
        return self.db.query(
            query,
            kwargs.get('params'),
            epoch=precision,
            expected_response_code=kwargs.get('expected_response_code') or 200,
            database=database,
        )

    def _write(self, points, database, retention_policy):
        """
        Write data points in the specified database.

        This method writes data points to the specified database. If the size of the data exceeds
        the limit of a UDP packet, it falls back to using a TCP connection for writing data.

        Args:
            points (list): The data points to be stored.
            database (str): The name of the database where the data points will be stored.
            retention_policy (str): The retention policy to be used for storing the data points.

        Returns:
            bool: True if the data points were successfully written, False otherwise.

        Raises:
            TimeseriesWriteException: If there is an error while writing the data points.
        """
        db = self.dbs['short'] if retention_policy else self.dbs['default']
        # If the size of data exceeds the limit of the UDP packet, then
        # fallback to use TCP connection for writing data.
        lines = make_lines({'points': points})
        if sys.getsizeof(lines) > 65000:
            # Size exceeds UDP limit, write using TCP.
            db = self.dbs['__all__']
        try:
            return db.write_points(
                points=lines.split('\n')[:-1],
                database=database,
                retention_policy=retention_policy,
                protocol='line',
            )
        except Exception as exception:
            logger.warning(f'got exception while writing to tsdb: {exception}')
            if isinstance(exception, self.client_error):
                exception_code = getattr(exception, 'code', None)
                # do not retry any request which returns 400
                if exception_code == 400:
                    return
            raise TimeseriesWriteException

    def _get_timestamp(self, timestamp=None):
        timestamp = timestamp or now()
        if isinstance(timestamp, datetime):
            return timestamp.isoformat(sep='T', timespec='microseconds')
        return timestamp

    def write(self, name, values, **kwargs):
        timestamp = self._get_timestamp(timestamp=kwargs.get('timestamp'))
        point = {
            'measurement': name,
            'tags': kwargs.get('tags'),
            'fields': values,
            'time': timestamp,
        }
        self._write(
            points=[point],
            database=kwargs.get('database') or self.db_name,
            retention_policy=kwargs.get('retention_policy'),
        )

    def batch_write(self, metric_data):
        data_points = {}
        for data in metric_data:
            org = data.get('database') or self.db_name
            retention_policy = data.get('retention_policy')
            if org not in data_points:
                data_points[org] = {}
            if retention_policy not in data_points[org]:
                data_points[org][retention_policy] = []
            timestamp = self._get_timestamp(timestamp=data.get('timestamp'))
            data_points[org][retention_policy].append(
                {
                    'measurement': data.get('name'),
                    'tags': data.get('tags'),
                    'fields': data.get('values'),
                    'time': timestamp,
                }
            )
        for database in data_points.keys():
            for rp in data_points[database].keys():
                self._write(
                    points=data_points[database][rp],
                    database=database,
                    retention_policy=rp,
                )

    def read(self, key, fields, tags, **kwargs):
        extra_fields = kwargs.get('extra_fields')
        since = kwargs.get('since')
        order = kwargs.get('order')
        limit = kwargs.get('limit')
        rp = kwargs.get('retention_policy')
        if extra_fields and extra_fields != '*':
            fields = ', '.join([fields] + extra_fields)
        elif extra_fields == '*':
            fields = '*'
        from_clause = f'{rp}.{key}' if rp else key
        q = f'SELECT {fields} FROM {from_clause}'
        conditions = []
        if since:
            conditions.append(f'time >= {since}')
        if tags:
            conditions.append(
                ' AND '.join(["{0} = '{1}'".format(*tag) for tag in tags.items()])
            )
        if conditions:
            conditions = 'WHERE %s' % ' AND '.join(conditions)
            q = f'{q} {conditions}'
        if order:
            # InfluxDB only allows ordering results by time
            if order == 'time':
                order = 'time ASC'
            elif order == '-time':
                order = 'time DESC'
            else:
                raise self.client_error(
                    f'Invalid order "{order}" passed.\nYou may pass "time" / "-time" to get '
                    'result sorted in ascending /descending order respectively.'
                )
            q = f'{q} ORDER BY {order}'
        if limit:
            q = f'{q} LIMIT {limit}'
        return list(self.query(q, precision='s').get_points())

    def get_list_query(self, query, precision='s'):
        result = self.query(query, precision=precision)
        if not len(result.keys()) or result.keys()[0][1] is None:
            return list(result.get_points())
        # Handles query which contains "GROUP BY TAG" clause
        result_points = {}
        for (measurement, tag), group_points in result.items():
            tag_suffix = '_'.join(tag.values())
            for point in group_points:
                values = {}
                for key, value in point.items():
                    if key != 'time':
                        values[tag_suffix] = value
                values['time'] = point['time']
                try:
                    result_points[values['time']].update(values)
                except KeyError:
                    result_points[values['time']] = values
        return list(result_points.values())

    @retry
    def get_list_retention_policies(self):
        return self.db.get_list_retention_policies()

    def delete_metric_data(self, key=None, tags=None):
        """Deletes a specific metric.

        Deletes a metric based on the key and tags provided, you may also
        choose to delete all metrics.
        """
        if not key and not tags:
            self.query('DROP SERIES FROM /.*/')
        else:
            self.delete_series(key, tags)

    @retry
    def delete_series(self, key=None, tags=None):
        self.db.delete_series(measurement=key, tags=tags)

    # Chart related functions below

    def validate_query(self, query):
        for word in self._FORBIDDEN:
            if word in query.lower():
                msg = _(f'the word "{word.upper()}" is not allowed')
                raise ValidationError({'configuration': msg})
        return self._is_aggregate(query)

    def _is_aggregate(self, q):
        q = q.upper()
        for word in self._AGGREGATE:
            if any(['%s(' % word in q, '|%s}' % word in q, '|%s|' % word in q]):
                return True
        return False

    def _clean_params(self, params):
        if params.get('end_date'):
            params['end_date'] = f'AND time <= \'{params["end_date"]}\''
        else:
            params['end_date'] = ''
        for key, value in params.items():
            if isinstance(value, list) or isinstance(value, tuple):
                params[key] = self._get_where_query(key, value)
        return params

    def _get_where_query(self, field, items):
        if not items:
            return ''
        lookup = []
        for item in items:
            lookup.append(f"{field} = '{item}'")
        return 'AND ({lookup})'.format(lookup=' OR '.join(lookup))

    def get_query(
        self,
        chart_type,
        params,
        time,
        group_map,
        summary=False,
        fields=None,
        query=None,
        timezone=settings.TIME_ZONE,
    ):
        query = self._fields(fields, query, params['field_name'])
        params = self._clean_params(params)
        query = query.format(**params)
        query = self._group_by(query, time, chart_type, group_map, strip=summary)
        if summary:
            query = f'{query} LIMIT 1'
        return f"{query} tz('{timezone}')"

    _group_by_time_tag_regex = re.compile(
        r'GROUP BY ((time\(\w+\))(?:,\s+\w+)?)', flags=re.IGNORECASE
    )
    _group_by_time_regex = re.compile(r'GROUP BY time\(\w+\)\s?', flags=re.IGNORECASE)
    _time_regex = re.compile(r'time\(\w+\)\s?', flags=re.IGNORECASE)
    _time_comma_regex = re.compile(r'time\(\w+\),\s?', flags=re.IGNORECASE)

    def _group_by(self, query, time, chart_type, group_map, strip=False):
        if not self.validate_query(query):
            return query
        if not strip and not chart_type == 'histogram':
            value = group_map[time]
            group_by = f'GROUP BY time({value})'
        else:
            # can be empty when getting summaries
            group_by = ''
        if 'GROUP BY' not in query.upper():
            query = f'{query} {group_by}'
        else:
            # The query could have GROUP BY clause for a TAG
            if group_by:
                # The query already contains "GROUP BY", therefore
                # we remove it from the "group_by" to avoid duplicating
                # "GROUP BY"
                group_by = group_by.replace('GROUP BY ', '')
                # We only need to substitute the time function.
                # The resulting query would be "GROUP BY time(<group_by>), <tag>"
                query = re.sub(self._time_regex, group_by, query)
            else:
                # The query should not include the "GROUP by time()"
                matches = re.search(self._group_by_time_tag_regex, query)
                group_by_fields = matches.group(1)
                if len(group_by_fields.split(',')) > 1:
                    # If the query has "GROUP BY time(), tag",
                    # then return "GROUP BY tag"
                    query = re.sub(self._time_comma_regex, '', query)
                else:
                    # If the query has only has "GROUP BY time()",
                    # then remove the "GROUP BY" clause
                    query = re.sub(self._group_by_time_regex, '', query)
        return query

    _fields_regex = re.compile(
        r'(?P<group>\{fields\|(?P<func>\w+)(?:\|(?P<op>.*?))?\})', flags=re.IGNORECASE
    )

    def _fields(self, fields, query, field_name):
        """Support substitution:

        of {fields|<FUNCTION_NAME>|<OPERATION>} with
        <FUNCTION_NAME>(field1) AS field1 <OPERATION>,
        <FUNCTION_NAME>(field2) AS field2 <OPERATION>
        """
        matches = re.search(self._fields_regex, query)
        if not matches and not fields:
            return query
        elif matches and not fields:
            groups = matches.groupdict()
            fields_key = groups.get('group')
            fields = [field_name]
        if fields and matches:
            groups = matches.groupdict()
            function = groups['func']  # required
            operation = groups.get('op')  # optional
            fields = [self.__transform_field(f, function, operation) for f in fields]
            fields_key = groups.get('group')
        else:
            fields_key = '{fields}'
        if fields:
            selected_fields = ', '.join(fields)
        return query.replace(fields_key, selected_fields)

    def __transform_field(self, field, function, operation=None):
        if operation:
            operation = f' {operation}'
        else:
            operation = ''
        return f'{function}("{field}"){operation} AS {field.replace("-", "_")}'

    def _get_top_fields(
        self,
        query,
        params,
        chart_type,
        group_map,
        number,
        time,
        timezone=settings.TIME_ZONE,
    ):
        q = self.get_query(
            query=query,
            params=params,
            chart_type=chart_type,
            group_map=group_map,
            summary=True,
            fields=['SUM(*)'],
            time=time,
            timezone=timezone,
        )
        res = list(self.query(q, precision='s').get_points())
        if not res:
            return []
        res = res[0]
        res = {key: value for key, value in res.items() if value is not None}
        sorted_dict = OrderedDict(sorted(res.items(), key=operator.itemgetter(1)))
        del sorted_dict['time']
        keys = list(sorted_dict.keys())
        keys.reverse()
        top = keys[0:number]
        return [item.replace('sum_', '') for item in top]



# flake8: noqa

RESULT_TCP = """
{
  "start": {
    "connected": [
      {
        "socket": 5,
        "local_host": "127.0.0.1",
        "local_port": 54966,
        "remote_host": "127.0.0.1",
        "remote_port": 5201
      }
    ],
    "version": "iperf 3.9",
    "system_info": "Linux openwisp-desktop 5.11.2-51-generic #58~20.04.1-Ubuntu SMP Tue Jun 14 11:29:12 UTC 2022 x86_64",
    "timestamp": {
      "time": "Thu, 30 Jun 2022 21:39:55 GMT",
      "timesecs": 1656625195
    },
    "connecting_to": {
      "host": "localhost",
      "port": 5201
    },
    "cookie": "npx4ad65t3j4wginxr4a7mqedmkhhspx3sob",
    "tcp_mss_default": 32768,
    "sock_bufsize": 0,
    "sndbuf_actual": 16384,
    "rcvbuf_actual": 131072,
    "test_start": {
      "protocol": "TCP",
      "num_streams": 1,
      "blksize": 131072,
      "omit": 0,
      "duration": 10,
      "bytes": 0,
      "blocks": 0,
      "reverse": 0,
      "tos": 0
    }
  },
  "intervals": [
    {
      "streams": [
        {
          "socket": 5,
          "start": 0,
          "end": 1.000048,
          "seconds": 1.000048041343689,
          "bytes": 5790760960,
          "bits_per_second": 46323862219.414116,
          "retransmits": 0,
          "snd_cwnd": 1506109,
          "rtt": 22,
          "rttvar": 3,
          "pmtu": 65535,
          "omitted": false,
          "sender": true
        }
      ],
      "sum": {
        "start": 0,
        "end": 1.000048,
        "seconds": 1.000048041343689,
        "bytes": 5790760960,
        "bits_per_second": 46323862219.414116,
        "retransmits": 0,
        "omitted": false,
        "sender": true
      }
    },
    {
      "streams": [
        {
          "socket": 5,
          "start": 1.000048,
          "end": 2.000185,
          "seconds": 1.0001369714736938,
          "bytes": 5463080960,
          "bits_per_second": 43698662209.83867,
          "retransmits": 0,
          "snd_cwnd": 2160939,
          "rtt": 22,
          "rttvar": 3,
          "pmtu": 65535,
          "omitted": false,
          "sender": true
        }
      ],
      "sum": {
        "start": 1.000048,
        "end": 2.000185,
        "seconds": 1.0001369714736938,
        "bytes": 5463080960,
        "bits_per_second": 43698662209.83867,
        "retransmits": 0,
        "omitted": false,
        "sender": true
      }
    },
    {
      "streams": [
        {
          "socket": 5,
          "start": 2.000185,
          "end": 3.00019,
          "seconds": 1.0000050067901611,
          "bytes": 5679349760,
          "bits_per_second": 45434570598.638954,
          "retransmits": 0,
          "snd_cwnd": 2553837,
          "rtt": 21,
          "rttvar": 1,
          "pmtu": 65535,
          "omitted": false,
          "sender": true
        }
      ],
      "sum": {
        "start": 2.000185,
        "end": 3.00019,
        "seconds": 1.0000050067901611,
        "bytes": 5679349760,
        "bits_per_second": 45434570598.638954,
        "retransmits": 0,
        "omitted": false,
        "sender": true
      }
    },
    {
      "streams": [
        {
          "socket": 5,
          "start": 3.00019,
          "end": 4.000232,
          "seconds": 1.0000419616699219,
          "bytes": 5710807040,
          "bits_per_second": 45684539320.4405,
          "retransmits": 0,
          "snd_cwnd": 2553837,
          "rtt": 24,
          "rttvar": 5,
          "pmtu": 65535,
          "omitted": false,
          "sender": true
        }
      ],
      "sum": {
        "start": 3.00019,
        "end": 4.000232,
        "seconds": 1.0000419616699219,
        "bytes": 5710807040,
        "bits_per_second": 45684539320.4405,
        "retransmits": 0,
        "omitted": false,
        "sender": true
      }
    },
    {
      "streams": [
        {
          "socket": 5,
          "start": 4.000232,
          "end": 5.000158,
          "seconds": 0.999925971031189,
          "bytes": 5307105280,
          "bits_per_second": 42459985508.942955,
          "retransmits": 0,
          "snd_cwnd": 3208667,
          "rtt": 27,
          "rttvar": 4,
          "pmtu": 65535,
          "omitted": false,
          "sender": true
        }
      ],
      "sum": {
        "start": 4.000232,
        "end": 5.000158,
        "seconds": 0.999925971031189,
        "bytes": 5307105280,
        "bits_per_second": 42459985508.942955,
        "retransmits": 0,
        "omitted": false,
        "sender": true
      }
    },
    {
      "streams": [
        {
          "socket": 5,
          "start": 5.000158,
          "end": 6.000229,
          "seconds": 1.0000710487365723,
          "bytes": 5308416000,
          "bits_per_second": 42464310964.35657,
          "retransmits": 0,
          "snd_cwnd": 3208667,
          "rtt": 28,
          "rttvar": 1,
          "pmtu": 65535,
          "omitted": false,
          "sender": true
        }
      ],
      "sum": {
        "start": 5.000158,
        "end": 6.000229,
        "seconds": 1.0000710487365723,
        "bytes": 5308416000,
        "bits_per_second": 42464310964.35657,
        "retransmits": 0,
        "omitted": false,
        "sender": true
      }
    },
    {
      "streams": [
        {
          "socket": 5,
          "start": 6.000229,
          "end": 7.000056,
          "seconds": 0.9998270273208618,
          "bytes": 5241569280,
          "bits_per_second": 41939808681.0701,
          "retransmits": 0,
          "snd_cwnd": 3208667,
          "rtt": 23,
          "rttvar": 4,
          "pmtu": 65535,
          "omitted": false,
          "sender": true
        }
      ],
      "sum": {
        "start": 6.000229,
        "end": 7.000056,
        "seconds": 0.9998270273208618,
        "bytes": 5241569280,
        "bits_per_second": 41939808681.0701,
        "retransmits": 0,
        "omitted": false,
        "sender": true
      }
    },
    {
      "streams": [
        {
          "socket": 5,
          "start": 7.000056,
          "end": 8.000202,
          "seconds": 1.0001460313797,
          "bytes": 5734400000,
          "bits_per_second": 45868501759.40331,
          "retransmits": 0,
          "snd_cwnd": 3208667,
          "rtt": 22,
          "rttvar": 1,
          "pmtu": 65535,
          "omitted": false,
          "sender": true
        }
      ],
      "sum": {
        "start": 7.000056,
        "end": 8.000202,
        "seconds": 1.0001460313797,
        "bytes": 5734400000,
        "bits_per_second": 45868501759.40331,
        "retransmits": 0,
        "omitted": false,
        "sender": true
      }
    },
    {
      "streams": [
        {
          "socket": 5,
          "start": 8.000202,
          "end": 9.0003,
          "seconds": 1.0000979900360107,
          "bytes": 5415895040,
          "bits_per_second": 43322915105.98867,
          "retransmits": 0,
          "snd_cwnd": 3208667,
          "rtt": 35,
          "rttvar": 12,
          "pmtu": 65535,
          "omitted": false,
          "sender": true
        }
      ],
      "sum": {
        "start": 8.000202,
        "end": 9.0003,
        "seconds": 1.0000979900360107,
        "bytes": 5415895040,
        "bits_per_second": 43322915105.98867,
        "retransmits": 0,
        "omitted": false,
        "sender": true
      }
    },
    {
      "streams": [
        {
          "socket": 5,
          "start": 9.0003,
          "end": 10.000218,
          "seconds": 0.999917984008789,
          "bytes": 5402787840,
          "bits_per_second": 43225847930.76398,
          "retransmits": 0,
          "snd_cwnd": 3208667,
          "rtt": 26,
          "rttvar": 17,
          "pmtu": 65535,
          "omitted": false,
          "sender": true
        }
      ],
      "sum": {
        "start": 9.0003,
        "end": 10.000218,
        "seconds": 0.999917984008789,
        "bytes": 5402787840,
        "bits_per_second": 43225847930.76398,
        "retransmits": 0,
        "omitted": false,
        "sender": true
      }
    }
  ],
  "end": {
    "streams": [
      {
        "sender": {
          "socket": 5,
          "start": 0,
          "end": 10.000218,
          "seconds": 10.000218,
          "bytes": 55054172160,
          "bits_per_second": 44042377604.16823,
          "retransmits": 0,
          "max_snd_cwnd": 3208667,
          "max_rtt": 35,
          "min_rtt": 21,
          "mean_rtt": 25,
          "sender": true
        },
        "receiver": {
          "socket": 5,
          "start": 0,
          "end": 10.000272,
          "seconds": 10.000218,
          "bytes": 55054172160,
          "bits_per_second": 44042139781.797935,
          "sender": true
        }
      }
    ],
    "sum_sent": {
      "start": 0,
      "end": 10.000218,
      "seconds": 10.000218,
      "bytes": 55054172160,
      "bits_per_second": 44042377604.16823,
      "retransmits": 0,
      "sender": true
    },
    "sum_received": {
      "start": 0,
      "end": 10.000272,
      "seconds": 10.000272,
      "bytes": 55054172160,
      "bits_per_second": 44042139781.797935,
      "sender": true
    },
    "cpu_utilization_percent": {
      "host_total": 99.49882081069975,
      "host_user": 0.6620490539150914,
      "host_system": 98.83676176238454,
      "remote_total": 0.377797593572381,
      "remote_user": 0.02174276147834767,
      "remote_system": 0.35605477540538377
    },
    "sender_tcp_congestion": "cubic",
    "receiver_tcp_congestion": "cubic"
  }
}
"""

RESULT_UDP = """
{
    "start": {
        "connected": [
            {
                "socket": 5,
                "local_host": "127.0.0.1",
                "local_port": 54477,
                "remote_host": "127.0.0.1",
                "remote_port": 5201
            }
        ],
        "version": "iperf 3.9",
        "system_info": "openwisp-desktop 5.11.2-51-generic #58~20.04.1-Ubuntu SMP Tue Jun 14 11:29:12 UTC 2022 x86_64",
        "timestamp": {
            "time": "Thu, 30 Jun 2022 21:10:31 GMT",
            "timesecs": 1656623431
        },
        "connecting_to": {
            "host": "localhost",
            "port": 5201
        },
        "cookie": "kvuxkz3ncutquvpl2evufmdkn726molzocot",
        "sock_bufsize": 0,
        "sndbuf_actual": 212992,
        "rcvbuf_actual": 212992,
        "test_start": {
            "protocol": "UDP",
            "num_streams": 1,
            "blksize": 32768,
            "omit": 0,
            "duration": 10,
            "bytes": 0,
            "blocks": 0,
            "reverse": 0,
            "tos": 0
        }
    },
    "intervals": [
        {
            "streams": [
                {
                    "socket": 5,
                    "start": 0,
                    "end": 1.000057,
                    "seconds": 1.0000569820404053,
                    "bytes": 131072,
                    "bits_per_second": 1048516.253404483,
                    "packets": 4,
                    "omitted": false,
                    "sender": true
                }
            ],
            "sum": {
                "start": 0,
                "end": 1.000057,
                "seconds": 1.0000569820404053,
                "bytes": 131072,
                "bits_per_second": 1048516.253404483,
                "packets": 4,
                "omitted": false,
                "sender": true
            }
        },
        {
            "streams": [
                {
                    "socket": 5,
                    "start": 1.000057,
                    "end": 2.000079,
                    "seconds": 1.000022053718567,
                    "bytes": 131072,
                    "bits_per_second": 1048552.875509981,
                    "packets": 4,
                    "omitted": false,
                    "sender": true
                }
            ],
            "sum": {
                "start": 1.000057,
                "end": 2.000079,
                "seconds": 1.000022053718567,
                "bytes": 131072,
                "bits_per_second": 1048552.875509981,
                "packets": 4,
                "omitted": false,
                "sender": true
            }
        },
        {
            "streams": [
                {
                    "socket": 5,
                    "start": 2.000079,
                    "end": 3.000079,
                    "seconds": 1,
                    "bytes": 131072,
                    "bits_per_second": 1048576,
                    "packets": 4,
                    "omitted": false,
                    "sender": true
                }
            ],
            "sum": {
                "start": 2.000079,
                "end": 3.000079,
                "seconds": 1,
                "bytes": 131072,
                "bits_per_second": 1048576,
                "packets": 4,
                "omitted": false,
                "sender": true
            }
        },
        {
            "streams": [
                {
                    "socket": 5,
                    "start": 3.000079,
                    "end": 4.000079,
                    "seconds": 1,
                    "bytes": 131072,
                    "bits_per_second": 1048576,
                    "packets": 4,
                    "omitted": false,
                    "sender": true
                }
            ],
            "sum": {
                "start": 3.000079,
                "end": 4.000079,
                "seconds": 1,
                "bytes": 131072,
                "bits_per_second": 1048576,
                "packets": 4,
                "omitted": false,
                "sender": true
            }
        },
        {
            "streams": [
                {
                    "socket": 5,
                    "start": 4.000079,
                    "end": 5.000182,
                    "seconds": 1.0001029968261719,
                    "bytes": 131072,
                    "bits_per_second": 1048468.0111225117,
                    "packets": 4,
                    "omitted": false,
                    "sender": true
                }
            ],
            "sum": {
                "start": 4.000079,
                "end": 5.000182,
                "seconds": 1.0001029968261719,
                "bytes": 131072,
                "bits_per_second": 1048468.0111225117,
                "packets": 4,
                "omitted": false,
                "sender": true
            }
        },
        {
            "streams": [
                {
                    "socket": 5,
                    "start": 5.000182,
                    "end": 6.000056,
                    "seconds": 0.9998739957809448,
                    "bytes": 131072,
                    "bits_per_second": 1048708.1416504055,
                    "packets": 4,
                    "omitted": false,
                    "sender": true
                }
            ],
            "sum": {
                "start": 5.000182,
                "end": 6.000056,
                "seconds": 0.9998739957809448,
                "bytes": 131072,
                "bits_per_second": 1048708.1416504055,
                "packets": 4,
                "omitted": false,
                "sender": true
            }
        },
        {
            "streams": [
                {
                    "socket": 5,
                    "start": 6.000056,
                    "end": 7.000056,
                    "seconds": 1,
                    "bytes": 131072,
                    "bits_per_second": 1048576,
                    "packets": 4,
                    "omitted": false,
                    "sender": true
                }
            ],
            "sum": {
                "start": 6.000056,
                "end": 7.000056,
                "seconds": 1,
                "bytes": 131072,
                "bits_per_second": 1048576,
                "packets": 4,
                "omitted": false,
                "sender": true
            }
        },
        {
            "streams": [
                {
                    "socket": 5,
                    "start": 7.000056,
                    "end": 8.000056,
                    "seconds": 1,
                    "bytes": 131072,
                    "bits_per_second": 1048576,
                    "packets": 4,
                    "omitted": false,
                    "sender": true
                }
            ],
            "sum": {
                "start": 7.000056,
                "end": 8.000056,
                "seconds": 1,
                "bytes": 131072,
                "bits_per_second": 1048576,
                "packets": 4,
                "omitted": false,
                "sender": true
            }
        },
        {
            "streams": [
                {
                    "socket": 5,
                    "start": 8.000056,
                    "end": 9.000057,
                    "seconds": 1.0000009536743164,
                    "bytes": 131072,
                    "bits_per_second": 1048575.0000009537,
                    "packets": 4,
                    "omitted": false,
                    "sender": true
                }
            ],
            "sum": {
                "start": 8.000056,
                "end": 9.000057,
                "seconds": 1.0000009536743164,
                "bytes": 131072,
                "bits_per_second": 1048575.0000009537,
                "packets": 4,
                "omitted": false,
                "sender": true
            }
        },
        {
            "streams": [
                {
                    "socket": 5,
                    "start": 9.000057,
                    "end": 10.00006,
                    "seconds": 1.0000029802322388,
                    "bytes": 131072,
                    "bits_per_second": 1048572.8750093132,
                    "packets": 4,
                    "omitted": false,
                    "sender": true
                }
            ],
            "sum": {
                "start": 9.000057,
                "end": 10.00006,
                "seconds": 1.0000029802322388,
                "bytes": 131072,
                "bits_per_second": 1048572.8750093132,
                "packets": 4,
                "omitted": false,
                "sender": true
            }
        }
    ],
    "end": {
        "streams": [
            {
                "udp": {
                    "socket": 5,
                    "start": 0,
                    "end": 10.00006,
                    "seconds": 10.00006,
                    "bytes": 1310720,
                    "bits_per_second": 1048569.7085817485,
                    "jitter_ms": 0.011259258240784126,
                    "lost_packets": 0,
                    "packets": 40,
                    "lost_percent": 0,
                    "out_of_order": 0,
                    "sender": true
                }
            }
        ],
        "sum": {
            "start": 0,
            "end": 10.000115,
            "seconds": 10.000115,
            "bytes": 1310720,
            "bits_per_second": 1048569.7085817485,
            "jitter_ms": 0.011259258240784126,
            "lost_packets": 0,
            "packets": 40,
            "lost_percent": 0,
            "sender": true
        },
        "cpu_utilization_percent": {
            "host_total": 0.6057128493969417,
            "host_user": 0,
            "host_system": 0.6057128493969417,
            "remote_total": 0.016163250220207454,
            "remote_user": 0.01616789349806445,
            "remote_system": 0
        }
    }
}
"""

RESULT_FAIL = """
{
    "start": {
        "connected": [],
        "version": "iperf 3.7",
        "system_info": "Linux vm-openwrt 4.14.171 #0 SMP Thu Feb 27 21:05:12 2020 x86_64"
    },
    "intervals": [],
    "end": {},
    "error": "error - unable to connect to server: Connection refused"
}
"""
RESULT_AUTH_FAIL = """
{
    "start": {
        "connected": [],
        "version": "iperf 3.7",
        "system_info": "Linux vm-openwrt 4.14.171 #0 SMP Thu Feb 27 21:05:12 2020 x86_64",
        "timestamp": {
            "time": "Tue, 19 Jul 2022 12:23:38 UTC",
            "timesecs": 1658233418
        },
        "connecting_to": {
            "host": "192.168.5.109",
            "port": 5201
        },
        "cookie": "llz5f6akwyonbtcj3fx4phvfaflohdlvxr4z",
        "tcp_mss_default": 1460
    },
    "intervals": [],
    "end": {},
    "error": "error - test authorization failed"
}
"""
PARAM_ERROR = """Usage: iperf3 [-s|-c host] [options]
       iperf3 [-h|--help] [-v|--version]

Server or Client:
  -p, --port      #         server port to listen on/connect to
  -f, --format   [kmgtKMGT] format to report: Kbits, Mbits, Gbits, Tbits
  -i, --interval  #         seconds between periodic throughput reports
  -F, --file name           xmit/recv the specified file
  -A, --affinity n/n,m      set CPU affinity
  -B, --bind      <host>    bind to the interface associated with the address <host>
  -V, --verbose             more detailed output
  -J, --json                output in JSON format
  --logfile f               send output to a log file
  --forceflush              force flushing output at every interval
  -d, --debug               emit debugging output
  -v, --version             show version information and quit
  -h, --help                show this message and quit
Server specific:
  -s, --server              run in server mode
  -D, --daemon              run the server as a daemon
  -I, --pidfile file        write PID file
  -1, --one-off             handle one client connection then exit
  --rsa-private-key-path    path to the RSA private key used to decrypt
                            authentication credentials
  --authorized-users-path   path to the configuration file containing user
                            credentials
Client specific:
  -c, --client    <host>    run in client mode, connecting to <host>
  -u, --udp                 use UDP rather than TCP
  --connect-timeout #       timeout for control connection setup (ms)
  -b, --bitrate #[KMG][/#]  target bitrate in bits/sec (0 for unlimited)
                            (default 1 Mbit/sec for UDP, unlimited for TCP)
                            (optional slash and packet count for burst mode)
  --pacing-timer #[KMG]     set the timing for pacing, in microseconds (default 1000)
  --fq-rate #[KMG]          enable fair-queuing based socket pacing in
                            bits/sec (Linux only)
  -t, --time      #         time in seconds to transmit for (default 10 secs)
  -n, --bytes     #[KMG]    number of bytes to transmit (instead of -t)
  -k, --blockcount #[KMG]   number of blocks (packets) to transmit (instead of -t or -n)
  -l, --length    #[KMG]    length of buffer to read or write
                            (default 128 KB for TCP, dynamic or 1460 for UDP)
  --cport         <port>    bind to a specific client port (TCP and UDP, default: ephemeral port)
  -P, --parallel  #         number of parallel client streams to run
  -R, --reverse             run in reverse mode (server sends, client receives)
  --bidir                   run in bidirectional mode.
                            Client and server send and receive data.
  -w, --window    #[KMG]    set window size / socket buffer size
  -C, --congestion <algo>   set TCP congestion control algorithm (Linux and FreeBSD only)
  -M, --set-mss   #         set TCP/SCTP maximum segment size (MTU - 40 bytes)
  -N, --no-delay            set TCP/SCTP no delay, disabling Nagle's Algorithm
  -4, --version4            only use IPv4
  -6, --version6            only use IPv6
  -S, --tos N               set the IP type of service, 0-255.
                            The usual prefixes for octal and hex can be used,
                            i.e. 52, 064 and 0x34 all specify the same value.
  --dscp N or --dscp val    set the IP dscp value, either 0-63 or symbolic.
                            Numeric values can be specified in decimal,
                            octal and hex (see --tos above).
  -L, --flowlabel N         set the IPv6 flow label (only supported on Linux)
  -Z, --zerocopy            use a 'zero copy' method of sending data
  -O, --omit N              omit the first n seconds
  -T, --title str           prefix every output line with this string
  --extra-data str          data string to include in client and server JSON
  --get-server-output       get results from server
  --udp-counters-64bit      use 64-bit counters in UDP test packets
  --repeating-payload       use repeating pattern in payload, instead of
                            randomized payload (like in iperf2)
  --username                username for authentication
  --rsa-public-key-path     path to the RSA public key used to encrypt
                            authentication credentials

[KMG] indicates options that support a K/M/G suffix for kilo-, mega-, or giga-

iperf3 homepage at: https://software.es.net/iperf/
Report bugs to:     https://github.com/esnet/iperf
iperf3: parameter error - you must specify username (max 20 chars), password (max 20 chars) and a path to a valid public rsa client to be used"""

TEST_RSA_KEY = """MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAwuEm+iYrfSWJOupy6X3N
dxZvUCxvmoL3uoGAs0O0Y32unUQrwcTIxudy38JSuCccD+k2Rf8S4WuZSiTxaoea
6Du99YQGVZeY67uJ21SWFqWU+w6ONUj3TrNNWoICN7BXGLE2BbSBz9YaXefE3aqw
GhEjQz364Itwm425vHn2MntSp0weWb4hUCjQUyyooRXPrFUGBOuY+VvAvMyAG4Uk
msapnWnBSxXt7Tbb++A5XbOMdM2mwNYDEtkD5ksC/x3EVBrI9FvENsH9+u/8J9Mf
2oPl4MnlCMY86MQypkeUn7eVWfDnseNky7TyC0/IgCXve/iaydCCFdkjyo1MTAA4
BQIDAQAB"""

INVALID_PARAMS = [
    {'host': ''},
    {'host': 12},
    {'host': 'test.openwisp.io'},
    {'username': 121},
    {'password': -323},
    {'rsa_public_key': 1334},
    {'username': ''},
    {'password': 0},
    {'rsa_public_key': 0},
    {
        'username': 'openwisp-test-user',
        'password': 'open-pass',
        'rsa_public_key': -1,
    },
    {
        'username': 1123,
        'password': 'rossi',
        'rsa_public_key': '',
    },
    {
        'username': 'openwisp-test-user',
        'password': -214,
    },
    {
        'client_options': {
            'port': 'testport',
            'time': 120,
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': '50M'},
        }
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 'testport',
            'time': 120,
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': '50M'},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 70000,
            'time': 120,
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': '50M'},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': -21,
            'time': 120,
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': '50M'},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'time': 1200000,
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': '50M'},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'time': 20,
            'tcp': {'bitrate': 10},
            'udp': {'bitrate': '50M'},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'time': 120,
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': 50},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'bytes': 20,
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': 50},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'bytes': '',
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': 50},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'bytes': -1,
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': 50},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'connect_timeout': -1,
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': 50},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'connect_timeout': '11000',
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': 50},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'blockcount': -13,
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': 50},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'blockcount': '',
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': 50},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'tcp': {'bitrate': '10M', 'length': 112},
            'udp': {'bitrate': 50},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'connect_timeout': 2000000,
            'blockcount': '100K',
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': 50, 'length': 9595},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'parallel': '12',
            'connect_timeout': 2000000,
            'blockcount': '100K',
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': 50, 'length': 9595},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'parallel': 0,
            'connect_timeout': 2000000,
            'blockcount': '100K',
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': 50, 'length': 9595},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'parallel': 250,
            'connect_timeout': 2000000,
            'blockcount': '100K',
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': 50, 'length': 9595},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'bidirectional': True,
            'connect_timeout': 2000000,
            'blockcount': '100K',
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': 50, 'length': 9595},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'reverse': False,
            'connect_timeout': 2000000,
            'blockcount': '100K',
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': 50, 'length': 9595},
        },
    },
    {
        'host': ['test.openwisp.io'],
        'client_options': {
            'port': 5201,
            'window': 0,
            'connect_timeout': 2000000,
            'blockcount': '100K',
            'tcp': {'bitrate': '10M'},
            'udp': {'bitrate': 50, 'length': 9595},
        },
    },
]



from import_export.fields import Field

from openwisp_controller.geo.exportable import GeoDeviceResource as BaseDeviceResource


class DeviceMonitoringResource(BaseDeviceResource):
    monitoring_status = Field(
        attribute='monitoring__status', column_name='monitoring_status', readonly=True
    )

    class Meta(BaseDeviceResource.Meta):
        fields = BaseDeviceResource.Meta.fields[:]  # copy
        # add monitoring status to the exportable fields of DeviceAdmin
        fields.insert(fields.index('config_status'), 'monitoring_status')
        export_order = fields


# this list of field is defined here to facilitate integartion testing
# with other modules which may need to mock this before the admin is loaded
_exportable_fields = DeviceMonitoringResource.Meta.fields



import uuid
from urllib.parse import urljoin

from django.contrib import admin
from django.contrib.contenttypes.admin import GenericStackedInline
from django.contrib.contenttypes.forms import BaseGenericInlineFormSet
from django.contrib.contenttypes.models import ContentType
from django.forms import ModelForm
from django.templatetags.static import static
from django.urls import resolve, reverse
from django.utils import timezone
from django.utils.formats import localize
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from import_export.admin import ImportExportMixin
from nested_admin.nested import (
    NestedGenericStackedInline,
    NestedModelAdmin,
    NestedStackedInline,
)
from swapper import load_model

from openwisp_controller.config.admin import DeactivatedDeviceReadOnlyMixin
from openwisp_controller.config.admin import DeviceAdmin as BaseDeviceAdmin
from openwisp_users.multitenancy import MultitenantAdminMixin
from openwisp_utils.admin import ReadOnlyAdmin

from ..monitoring.admin import MetricAdmin
from ..settings import MONITORING_API_BASEURL, MONITORING_API_URLCONF
from . import settings as app_settings
from .exportable import DeviceMonitoringResource
from .filters import DeviceFilter, DeviceGroupFilter, DeviceOrganizationFilter

DeviceData = load_model('device_monitoring', 'DeviceData')
WifiSession = load_model('device_monitoring', 'WifiSession')
DeviceMonitoring = load_model('device_monitoring', 'DeviceMonitoring')
AlertSettings = load_model('monitoring', 'AlertSettings')
Chart = load_model('monitoring', 'Chart')
Device = load_model('config', 'Device')
Metric = load_model('monitoring', 'Metric')
Notification = load_model('openwisp_notifications', 'Notification')
Check = load_model('check', 'Check')
Organization = load_model('openwisp_users', 'Organization')


class CheckInlineFormSet(BaseGenericInlineFormSet):
    def full_clean(self):
        for form in self.forms:
            obj = form.instance
            if not obj.content_type or not obj.object_id:
                setattr(
                    obj,
                    self.ct_field.get_attname(),
                    ContentType.objects.get_for_model(self.instance).pk,
                )
                setattr(obj, self.ct_fk_field.get_attname(), self.instance.pk)
        super().full_clean()


class InlinePermissionMixin:
    """
    Manages permissions for the inline objects.

    This mixin class handles the permissions for the inline objects.
    It checks if the user has permission to modify the main object
    (e.g., view_model, add_model, change_model, or delete_model),
    and if not, it checks if the user has permission to modify the
    inline object (e.g., view_model_inline, add_model_inline,
    change_model_inline, or delete_model_inline).
    """

    def has_add_permission(self, request, obj=None):
        return super(admin.options.InlineModelAdmin, self).has_add_permission(
            request
        ) or request.user.has_perm(
            f'{self.model._meta.app_label}.add_{self.inline_permission_suffix}'
        )

    def has_change_permission(self, request, obj=None):
        return super(admin.options.InlineModelAdmin, self).has_change_permission(
            request, obj
        ) or request.user.has_perm(
            f'{self.model._meta.app_label}.change_{self.inline_permission_suffix}'
        )

    def has_view_permission(self, request, obj=None):
        return super(admin.options.InlineModelAdmin, self).has_view_permission(
            request, obj
        ) or request.user.has_perm(
            f'{self.model._meta.app_label}.view_{self.inline_permission_suffix}'
        )

    def has_delete_permission(self, request, obj=None):
        return super(admin.options.InlineModelAdmin, self).has_delete_permission(
            request, obj
        ) or request.user.has_perm(
            f'{self.model._meta.app_label}.delete_{self.inline_permission_suffix}'
        )


class DeactivatedDeviceReadOnlyInlinePermissionMixin(
    DeactivatedDeviceReadOnlyMixin,
    InlinePermissionMixin,
):
    def has_add_permission(self, request, obj=None):
        perm = super().has_add_permission(request, obj)
        return self._has_permission(request, obj, perm)

    def has_change_permission(self, request, obj=None):
        perm = super().has_change_permission(request, obj)
        return self._has_permission(request, obj, perm)

    def has_view_permission(self, request, obj=None):
        perm = super().has_view_permission(request, obj)
        return self._has_permission(request, obj, perm)

    def has_delete_permission(self, request, obj=None):
        perm = super().has_delete_permission(request, obj)
        return self._has_permission(request, obj, perm)


class CheckInline(DeactivatedDeviceReadOnlyInlinePermissionMixin, GenericStackedInline):
    model = Check
    extra = 0
    formset = CheckInlineFormSet
    fields = [
        'is_active',
        'check_type',
    ]
    inline_permission_suffix = 'check_inline'

    def get_fields(self, request, obj=None):
        if not self.has_change_permission(request, obj) or not self.has_view_permission(
            request, obj
        ):
            return ['check_type', 'is_active']
        return super().get_fields(request, obj)

    def get_readonly_fields(self, request, obj=None):
        if not self.has_change_permission(request, obj) or not self.has_view_permission(
            request, obj
        ):
            return ['check_type']
        return super().get_readonly_fields(request, obj)


class AlertSettingsForm(ModelForm):
    def __init__(self, *args, **kwargs):
        instance = kwargs.get('instance')
        if instance:
            kwargs['initial'] = {
                'custom_tolerance': instance.tolerance,
                'custom_operator': instance.operator,
                'custom_threshold': instance.threshold,
            }
        super().__init__(*args, **kwargs)

    def _post_clean(self):
        self.instance._delete_instance = False
        if all(
            self.cleaned_data[field] is None
            for field in [
                'custom_operator',
                'custom_threshold',
                'custom_tolerance',
            ]
        ):
            # "_delete_instance" flag signifies that
            # the fields have been set to None by the
            # user. Hence, the object should be deleted.
            self.instance._delete_instance = True
        super()._post_clean()

    def save(self, commit=True):
        if self.instance._delete_instance:
            self.instance.delete()
            return self.instance
        return super().save(commit)


class AlertSettingsInline(InlinePermissionMixin, NestedStackedInline):
    model = AlertSettings
    extra = 1
    max_num = 1
    exclude = ['created', 'modified']
    form = AlertSettingsForm
    inline_permission_suffix = 'alertsettings_inline'

    def get_queryset(self, request):
        return super().get_queryset(request).order_by('created')


class MetricInline(
    DeactivatedDeviceReadOnlyInlinePermissionMixin, NestedGenericStackedInline
):
    model = Metric
    extra = 0
    inlines = [AlertSettingsInline]
    fieldsets = [
        (
            None,
            {
                'fields': (
                    'name',
                    'is_healthy',
                )
            },
        ),
        (
            _('Advanced options'),
            {'classes': ('collapse',), 'fields': ('field_name',)},
        ),
    ]

    readonly_fields = ['name', 'is_healthy']
    # Explicitly changed name from Metrics to Alert Settings
    verbose_name = _('Alert Settings')
    verbose_name_plural = verbose_name
    inline_permission_suffix = 'alertsettings_inline'
    # Ordering queryset by metric name
    ordering = ('name',)

    def get_fieldsets(self, request, obj=None):
        if not self.has_change_permission(request, obj) or not self.has_view_permission(
            request, obj
        ):
            return [
                (None, {'fields': ('is_healthy',)}),
            ]
        return super().get_fieldsets(request, obj)

    def get_queryset(self, request):
        # Only show 'Metrics' that have 'AlertSettings' objects
        return super().get_queryset(request).filter(alertsettings__isnull=False)

    def has_add_permission(self, request, obj=None):
        # We need to restrict the users from adding the 'metrics' since
        # they're created by the system automatically with default 'alertsettings'
        return False

    def has_delete_permission(self, request, obj=None):
        # We need to restrict the users from deleting the 'metrics' since
        # they're created by the system automatically with default 'alertsettings'
        return False


class DeviceAdmin(BaseDeviceAdmin, NestedModelAdmin):
    change_form_template = 'admin/monitoring/device/change_form.html'
    list_filter = ['monitoring__status'] + BaseDeviceAdmin.list_filter
    list_select_related = ['monitoring'] + list(BaseDeviceAdmin.list_select_related)
    list_display = list(BaseDeviceAdmin.list_display)
    list_display.insert(list_display.index('config_status'), 'health_status')
    readonly_fields = ['health_status'] + BaseDeviceAdmin.readonly_fields

    class Media:
        js = (
            tuple(BaseDeviceAdmin.Media.js)
            + (
                'monitoring/js/lib/percircle.min.js',
                'monitoring/js/alert-settings.js',
            )
            + MetricAdmin.Media.js
            + ('monitoring/js/chart-utils.js',)
            + ('monitoring/js/lib/moment.min.js',)
            + ('monitoring/js/lib/daterangepicker.min.js',)
        )
        css = {
            'all': (
                'monitoring/css/percircle.min.css',
                'monitoring/css/daterangepicker.css',
            )
            + MetricAdmin.Media.css['all']
        }

    def get_extra_context(self, pk=None):
        ctx = super().get_extra_context(pk)
        if pk:
            device_data = DeviceData(pk=uuid.UUID(pk))
            api_url = reverse(
                'monitoring:api_device_metric',
                urlconf=MONITORING_API_URLCONF,
                args=[pk],
            )
            if MONITORING_API_BASEURL:
                api_url = urljoin(MONITORING_API_BASEURL, api_url)
            ctx.update(
                {
                    'device_data': device_data.data_user_friendly,
                    'api_url': api_url,
                    'default_time': Chart.DEFAULT_TIME,
                    'MAC_VENDOR_DETECTION': app_settings.MAC_VENDOR_DETECTION,
                }
            )
        return ctx

    def health_checks(self, obj):
        metric_rows = []
        for metric in DeviceData(pk=obj.pk).metrics.filter(alertsettings__isnull=False):
            health = 'yes' if metric.is_healthy else 'no'
            icon_url = static(f'admin/img/icon-{health}.svg')
            metric_rows.append(
                f'<li><img src="{icon_url}" ' f'alt="health"> {metric.name}</li>'
            )
        return format_html(
            mark_safe(f'<ul class="health_checks">{"".join(metric_rows)}</ul>')
        )

    health_checks.short_description = _('health checks')

    def health_status(self, obj):
        return format_html(
            mark_safe('<span class="health-{0}">{1}</span>'),
            obj.monitoring.status,
            obj.monitoring.get_status_display(),
        )

    health_status.short_description = _('health status')

    def get_object(self, request, object_id, from_field=None):
        obj = super().get_object(request, object_id, from_field=from_field)
        if obj and obj.wifisession_set.exists():
            # We need to provide default formset values
            # to avoid management formset errors when wifi sessions
            # are created while editing the DeviceAdmin change page
            wifisession_formset_data = {
                'wifisession_set-TOTAL_FORMS': '1',
                'wifisession_set-INITIAL_FORMS': '1',
            }
            request.POST = request.POST.copy()
            request.POST.update(wifisession_formset_data)
        return obj

    def get_form(self, request, obj=None, **kwargs):
        """Adds the help_text of DeviceMonitoring.status field"""
        health_status = DeviceMonitoring._meta.get_field('status').help_text
        kwargs.update(
            {'help_texts': {'health_status': health_status.replace('\n', '<br>')}}
        )
        return super().get_form(request, obj, **kwargs)

    def get_fields(self, request, obj=None):
        fields = list(super().get_fields(request, obj))
        if obj and not obj._state.adding:
            fields.insert(fields.index('last_ip'), 'health_status')
        if not obj or obj.monitoring.status in ['ok', 'unknown', 'deactivated']:
            return fields
        fields.insert(fields.index('health_status') + 1, 'health_checks')
        return fields

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = super().get_readonly_fields(request, obj)
        if not obj or obj.monitoring.status in ['ok', 'unknown']:
            return readonly_fields
        readonly_fields = list(readonly_fields)
        readonly_fields.append('health_checks')
        return readonly_fields

    def get_inlines(self, request, obj=None):
        inlines = super().get_inlines(request, obj)
        inlines = list(inlines + [CheckInline, MetricInline])
        # This attribute needs to be set for nested inline
        for inline in inlines:
            if not hasattr(inline, 'sortable_options'):
                inline.sortable_options = {'disabled': True}
        if not obj or obj._state.adding or obj.organization.is_active is False:
            inlines.remove(CheckInline)
            inlines.remove(MetricInline)
        return inlines


class DeviceAdminExportable(ImportExportMixin, DeviceAdmin):
    resource_class = DeviceMonitoringResource
    # Added to support both reversion and import-export
    change_list_template = 'admin/config/change_list_device.html'


class WifiSessionAdminHelperMixin:
    def _get_boolean_html(self, value):
        icon_type = 'unknown'
        if value is True:
            icon_type = 'yes'
        elif value is False:
            icon_type = 'no'
        icon = static(f'admin/img/icon-{icon_type}.svg')
        return mark_safe(f'<img src="{icon}">')

    def he(self, obj):
        return self._get_boolean_html(obj.wifi_client.he)

    he.short_description = 'WiFi 6 (802.11ax)'

    def vht(self, obj):
        return self._get_boolean_html(obj.wifi_client.vht)

    vht.short_description = 'WiFi 5 (802.11ac)'

    def ht(self, obj):
        return self._get_boolean_html(obj.wifi_client.ht)

    ht.short_description = 'WiFi 4 (802.11n)'

    def wmm(self, obj):
        return self._get_boolean_html(obj.wifi_client.wmm)

    wmm.short_description = 'WMM'

    def wds(self, obj):
        return self._get_boolean_html(obj.wifi_client.wds)

    wds.short_description = 'WDS'

    def wps(self, obj):
        return self._get_boolean_html(obj.wifi_client.wps)

    wps.short_description = 'WPS'

    def get_stop_time(self, obj):
        if obj.stop_time is None:
            return mark_safe('<strong style="color:green;">online</strong>')
        return localize(timezone.localtime(obj.stop_time))

    get_stop_time.short_description = _('stop time')

    def related_device(self, obj):
        app_label = Device._meta.app_label
        url = reverse(f'admin:{app_label}_device_change', args=[obj.device_id])
        return mark_safe(f'<a href="{url}">{obj.device}</a>')

    related_device.short_description = _('device')

    def related_organization(self, obj):
        app_label = Organization._meta.app_label
        url = reverse(
            f'admin:{app_label}_organization_change', args=[obj.device.organization_id]
        )
        return mark_safe(f'<a href="{url}">{obj.device.organization}</a>')

    related_organization.short_description = _('organization')


class WiFiSessionInline(WifiSessionAdminHelperMixin, admin.TabularInline):
    model = WifiSession
    fk_name = 'device'
    fields = [
        'get_mac_address',
        'vendor',
        'ssid',
        'interface_name',
        'he',
        'vht',
        'ht',
        'start_time',
        'get_stop_time',
    ]
    readonly_fields = fields
    can_delete = False
    extra = 0
    template = 'admin/monitoring/device/wifisession_tabular.html'

    class Media:
        css = {'all': ('monitoring/css/wifi-sessions.css',)}
        js = ['admin/js/jquery.init.js', 'monitoring/js/wifi-session-inline.js']

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def get_queryset(self, request, select_related=True):
        qs = super().get_queryset(request).filter(stop_time__isnull=True)
        resolved = resolve(request.path_info)
        if 'object_id' in resolved.kwargs:
            qs = qs.filter(device_id=resolved.kwargs['object_id'])
        if select_related:
            qs = qs.select_related('wifi_client')
        return qs

    def _get_conditional_queryset(self, request, obj, select_related=False):
        return self.get_queryset(request, select_related=select_related).exists()

    def get_mac_address(self, obj):
        app_label = WifiSession._meta.app_label
        url = reverse(f'admin:{app_label}_wifisession_change', args=[obj.id])
        return mark_safe(f'<a href="{url}">{obj.mac_address}</a>')

    get_mac_address.short_description = _('MAC address')


class WifiSessionAdmin(
    WifiSessionAdminHelperMixin, MultitenantAdminMixin, ReadOnlyAdmin
):
    multitenant_parent = 'device'
    model = WifiSession
    list_display = [
        'mac_address',
        'vendor',
        'related_organization',
        'related_device',
        'ssid',
        'he',
        'vht',
        'ht',
        'start_time',
        'get_stop_time',
    ]
    fields = [
        'related_organization',
        'mac_address',
        'vendor',
        'related_device',
        'ssid',
        'interface_name',
        'he',
        'vht',
        'ht',
        'wmm',
        'wds',
        'wps',
        'start_time',
        'get_stop_time',
        'modified',
    ]
    search_fields = ['wifi_client__mac_address', 'device__name', 'device__mac_address']
    list_filter = [
        DeviceOrganizationFilter,
        DeviceFilter,
        DeviceGroupFilter,
        'start_time',
        'stop_time',
    ]

    def get_readonly_fields(self, request, obj=None):
        fields = super().get_readonly_fields(request, obj)
        fields += [
            'related_organization',
            'mac_address',
            'vendor',
            'he',
            'vht',
            'ht',
            'wmm',
            'wds',
            'wps',
            'get_stop_time',
            'modified',
            'related_device',
        ]
        return fields

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related(
                'wifi_client', 'device', 'device__organization', 'device__group'
            )
        )

    def has_delete_permission(self, request, obj=None):
        return super(admin.ModelAdmin, self).has_delete_permission(request, obj)


admin.site.unregister(Device)
admin.site.register(Device, DeviceAdminExportable)

if app_settings.WIFI_SESSIONS_ENABLED:
    admin.site.register(WifiSession, WifiSessionAdmin)
    DeviceAdmin.conditional_inlines.append(WiFiSessionInline)



chart_query = {
    'uptime': {
        'influxdb': (
            "SELECT MEAN({field_name})*100 AS uptime FROM {key} WHERE "
            "time >= '{time}' {end_date} AND content_type = '{content_type}' AND "
            "object_id = '{object_id}' GROUP BY time(1d)"
        )
    },
    'packet_loss': {
        'influxdb': (
            "SELECT MEAN(loss) AS packet_loss FROM {key} WHERE "
            "time >= '{time}' {end_date} AND content_type = '{content_type}' AND "
            "object_id = '{object_id}' GROUP BY time(1d)"
        )
    },
    'rtt': {
        'influxdb': (
            "SELECT MEAN(rtt_avg) AS RTT_average, MEAN(rtt_max) AS "
            "RTT_max, MEAN(rtt_min) AS RTT_min FROM {key} WHERE "
            "time >= '{time}' {end_date} AND content_type = '{content_type}' AND "
            "object_id = '{object_id}' GROUP BY time(1d)"
        )
    },
    'wifi_clients': {
        'influxdb': (
            "SELECT COUNT(DISTINCT({field_name})) AS wifi_clients FROM {key} "
            "WHERE time >= '{time}' {end_date} AND content_type = '{content_type}' "
            "AND object_id = '{object_id}' AND ifname = '{ifname}' "
            "GROUP BY time(1d)"
        )
    },
    'general_wifi_clients': {
        'influxdb': (
            "SELECT COUNT(DISTINCT({field_name})) AS wifi_clients FROM {key} "
            "WHERE time >= '{time}' {end_date} {organization_id} {location_id} {floorplan_id} "
            "GROUP BY time(1d)"
        )
    },
    'traffic': {
        'influxdb': (
            "SELECT SUM(tx_bytes) / 1000000000 AS upload, "
            "SUM(rx_bytes) / 1000000000 AS download FROM {key} "
            "WHERE time >= '{time}' {end_date} AND content_type = '{content_type}' "
            "AND object_id = '{object_id}' AND ifname = '{ifname}' "
            "GROUP BY time(1d)"
        )
    },
    'general_traffic': {
        'influxdb': (
            "SELECT SUM(tx_bytes) / 1000000000 AS upload, "
            "SUM(rx_bytes) / 1000000000 AS download FROM {key} "
            "WHERE time >= '{time}' {end_date} {organization_id} {location_id} "
            "{floorplan_id} {ifname} "
            "GROUP BY time(1d)"
        )
    },
    'memory': {
        'influxdb': (
            "SELECT MEAN(percent_used) AS memory_usage "
            "FROM {key} WHERE time >= '{time}' {end_date} AND content_type = '{content_type}' "
            "AND object_id = '{object_id}' GROUP BY time(1d)"
        )
    },
    'cpu': {
        'influxdb': (
            "SELECT MEAN(cpu_usage) AS CPU_load FROM {key} WHERE "
            "time >= '{time}' {end_date} AND content_type = '{content_type}' AND "
            "object_id = '{object_id}' GROUP BY time(1d)"
        )
    },
    'disk': {
        'influxdb': (
            "SELECT MEAN(used_disk) AS disk_usage FROM {key} WHERE "
            "time >= '{time}' {end_date} AND content_type = '{content_type}' AND "
            "object_id = '{object_id}' GROUP BY time(1d)"
        )
    },
    'signal_strength': {
        'influxdb': (
            "SELECT ROUND(MEAN(signal_strength)) AS signal_strength, "
            "ROUND(MEAN(signal_power)) AS signal_power FROM {key} WHERE "
            "time >= '{time}' {end_date} AND content_type = '{content_type}' AND "
            "object_id = '{object_id}' GROUP BY time(1d)"
        )
    },
    'signal_quality': {
        'influxdb': (
            "SELECT ROUND(MEAN(signal_quality)) AS signal_quality, "
            "ROUND(MEAN(snr)) AS signal_to_noise_ratio FROM {key} WHERE "
            "time >= '{time}' {end_date} AND content_type = '{content_type}' AND "
            "object_id = '{object_id}' GROUP BY time(1d)"
        )
    },
    'access_tech': {
        'influxdb': (
            "SELECT MODE(access_tech) AS access_tech FROM {key} WHERE "
            "time >= '{time}' {end_date} AND content_type = '{content_type}' AND "
            "object_id = '{object_id}' GROUP BY time(1d)"
        )
    },
    'bandwidth': {
        'influxdb': (
            "SELECT MEAN(sent_bps_tcp) / 1000000000 AS TCP, "
            "MEAN(sent_bps_udp) / 1000000000 AS UDP FROM {key} WHERE "
            "time >= '{time}' AND content_type = '{content_type}' AND "
            "object_id = '{object_id}' GROUP BY time(1d)"
        )
    },
    'transfer': {
        'influxdb': (
            "SELECT SUM(sent_bytes_tcp) / 1000000000 AS TCP,"
            "SUM(sent_bytes_udp) / 1000000000 AS UDP FROM {key} WHERE "
            "time >= '{time}' AND content_type = '{content_type}' AND "
            "object_id = '{object_id}' GROUP BY time(1d)"
        )
    },
    'retransmits': {
        'influxdb': (
            "SELECT MEAN(retransmits) AS retransmits FROM {key} "
            "WHERE time >= '{time}' AND content_type = '{content_type}' "
            "AND object_id = '{object_id}' GROUP BY time(1d)"
        )
    },
    'jitter': {
        'influxdb': (
            "SELECT MEAN(jitter) AS jitter FROM {key} "
            "WHERE time >= '{time}' AND content_type = '{content_type}' "
            "AND object_id = '{object_id}' GROUP BY time(1d)"
        )
    },
    'datagram': {
        'influxdb': (
            "SELECT MEAN(lost_packets) AS lost_datagram,"
            "MEAN(total_packets) AS total_datagram FROM {key} WHERE "
            "time >= '{time}' AND content_type = '{content_type}' "
            "AND object_id = '{object_id}' GROUP BY time(1d)"
        )
    },
    'datagram_loss': {
        'influxdb': (
            "SELECT MEAN(lost_percent) AS datagram_loss FROM {key} "
            "WHERE time >= '{time}' AND content_type = '{content_type}' "
            "AND object_id = '{object_id}' GROUP BY time(1d)"
        )
    },
}

default_chart_query = [
    "SELECT {field_name} FROM {key} WHERE time >= '{time}' {end_date}",
    " AND content_type = '{content_type}' AND object_id = '{object_id}'",
]

device_data_query = (
    "SELECT data FROM {0}.{1} WHERE pk = '{2}' " "ORDER BY time DESC LIMIT 1"
)
